"""Abstract sample operators."""

from __future__ import annotations

from abc import abstractmethod
from math import inf
from typing import Generic, Literal

import pydantic

from ..dataflow import SampleProcessStatus
from ..enums import OperatorType
from ..models import FeatureType, RoiType, Sample
from ..storage import SampleStorage
from .base import BaseOperator


class SampleOperator(BaseOperator[SampleProcessStatus], Generic[RoiType, FeatureType]):
    """Base operator for sample storage."""

    type: Literal[OperatorType.SAMPLE] = OperatorType.SAMPLE

    def apply(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        """Apply the operator function to the data."""
        self.check_status(data.get_status())

        if hasattr(self, "pre_apply"):
            self.pre_apply()  # type: ignore

        self._apply_operator(data)

        if hasattr(self, "post_apply"):
            self.post_apply()  # type: ignore

        self.update_status(data.get_status())

    @abstractmethod
    def _apply_operator(self, data: SampleStorage) -> None: ...


class RoiExtractor(SampleOperator[RoiType, FeatureType]):
    """Extract ROIs from raw sample data."""

    def _apply_operator(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        sample = data.get_sample()
        rois = self.extract_rois(sample)
        data.add_rois(*rois)

    @abstractmethod
    def extract_rois(self, sample: Sample) -> list[RoiType]: ...  # noqa

    def get_expected_status_in(self) -> SampleProcessStatus:
        """Get the expected status before performing ROI extraction."""
        return SampleProcessStatus()

    def get_expected_status_out(self) -> SampleProcessStatus:
        """Get the expected status after performing ROI extraction."""
        return SampleProcessStatus(roi_extracted=True)


class RoiTransformer(SampleOperator[RoiType, FeatureType]):
    """Transform ROIs from raw sample data.

    Must implement the `transform_roi` method, which takes a single ROI and transform it inplace.

    If `transform_roi` returns ``None``, the ROI is removed from the sample storage.

    """

    def _apply_operator(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        rm_roi_list = list()
        for roi in data.list_rois():
            transformed = self.transform_roi(roi)
            if transformed is None:
                rm_roi_list.append(roi.id)
        data.delete_rois(*rm_roi_list)

    @abstractmethod
    def transform_roi(self, roi: RoiType) -> RoiType | None: ...  # noqa

    def get_expected_status_in(self) -> SampleProcessStatus:
        """Get the expected status before performing ROI extraction."""
        return SampleProcessStatus(roi_extracted=True)

    def get_expected_status_out(self) -> SampleProcessStatus:
        """Get the expected status after performing ROI extraction."""
        return SampleProcessStatus(roi_extracted=True)


class FeatureExtractor(SampleOperator[RoiType, FeatureType]):
    """Extract features from ROIs.

    Must implement the `extract_rois` method, which takes a single ROI and creates a list of features.

    Provides descriptor based filtering of features using the :py:attr:`bounds`.

    """

    bounds: dict[str, tuple[pydantic.NonNegativeFloat | None, pydantic.NonNegativeFloat | None]] = dict()
    """Define valid boundaries for each feature descriptor. Boundaries are expressed by mapping descriptor
    names to a tuple lower and upper bounds. If only a lower/upper bound is required, ``None`` must be used
    (e.g. ``(None, 10.0)`` to use only an upper bound)."""

    def _apply_operator(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        bounds = dict()
        for descriptor, (lower, upper) in self.bounds.items():
            lower = lower if lower is not None else -inf
            upper = upper if upper is not None else inf
            bounds[descriptor] = lower, upper

        features: list[FeatureType] = list()
        for roi in data.list_rois():
            extracted = [x for x in self.extract_features(roi) if x.has_descriptors_in_range(**bounds)]
            features.extend(extracted)
        data.add_features(*features)

    @abstractmethod
    def extract_features(self, roi: RoiType) -> list[FeatureType]: ...  # noqa

    def get_expected_status_in(self) -> SampleProcessStatus:
        """Get the expected status before performing ROI extraction."""
        return SampleProcessStatus(roi_extracted=True)

    def get_expected_status_out(self) -> SampleProcessStatus:
        """Get the expected status after performing ROI extraction."""
        return SampleProcessStatus(roi_extracted=True, feature_extracted=True)


class FeatureTransformer(SampleOperator[RoiType, FeatureType]):
    """Apply a transformation to individual features.

    Must implement the `transform_feature` method, which takes a single feature and transform it inplace.

    """

    def _apply_operator(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        rm_features = list()
        for feature in data.list_features():
            transformed = self.transform_feature(feature)
            if transformed is None:
                rm_features.append(feature.id)
        data.delete_features(*rm_features)

    @abstractmethod
    def transform_feature(self, feature: FeatureType) -> FeatureType | None: ...  # noqa

    def get_expected_status_in(self) -> SampleProcessStatus:
        """Get the expected status before performing ROI extraction."""
        return SampleProcessStatus(roi_extracted=True, feature_extracted=True)

    def get_expected_status_out(self) -> SampleProcessStatus:
        """Get the expected status after performing ROI extraction."""
        return SampleProcessStatus(roi_extracted=True, feature_extracted=True)



import numpy as np
import matplotlib.pyplot as plt

from tidyms2.annotation.correspondence import cluster_dbscan, split_cluster_gmm

np.random.seed(1234)
n = 200
X1 = np.random.normal(size=(n, 2))
samples = np.hstack((np.arange(n), np.arange(n)))
X2 = np.random.normal(size=(n, 2), loc=(2, 2))
X = np.vstack((X1, X2))

dbscan_labels = cluster_dbscan(X, 2.0, 50, 10000)
gmm_labels, score = split_cluster_gmm(X, samples, 2, 3.0)

fig, ax = plt.subplots()
for l in np.unique(gmm_labels):
    ax.scatter(*X[gmm_labels == l].T, label=l)

ax.set_xlabel("m/z")
ax.set_ylabel("Rt")
ax.legend(title="GMM labels")


import numpy as np
import pytest

from tidyms2.core.enums import NormalizationMethod, ScalingMethod
from tidyms2.core.utils import transformation


class TestNormalizeSum:
    def test_non_2d_array_raises_error(self):
        X = np.arange(5)
        with pytest.raises(ValueError):
            transformation.normalize(X)

    def test_2d_empty_array_raise_error(self):
        X = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            transformation.normalize(X)

    def test_single_column_single_use_str_method(self):
        X = np.array([[10.0]])

        actual = transformation.normalize(X, method="sum")
        expected = np.array([[1.0]])

        assert np.allclose(actual, expected)

    def test_single_column_single_row_returns_one(self):
        X = np.array([[10.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.SUM)
        expected = np.array([[1.0]])

        assert np.allclose(actual, expected)

    def test_single_single_row_returns_one(self):
        X = np.array([[1.0, 2.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.SUM)
        assert X.shape == actual.shape
        row_sum = np.nansum(actual, axis=1, keepdims=True)
        assert row_sum.shape == (X.shape[0], 1)
        assert np.allclose(row_sum, 1.0)

    def test_single_single_row_with_nans_returns_one(self):
        X = np.array([[1.0, 2.0, np.nan]])

        actual = transformation.normalize(X, method=NormalizationMethod.SUM)
        assert X.shape == actual.shape
        row_sum = np.nansum(actual, axis=1, keepdims=True)
        assert row_sum.shape == (X.shape[0], 1)
        assert np.allclose(row_sum, 1.0)

    def test_single_single_row_with_all_nans_return_all_nan(self):
        X = np.array([[np.nan, np.nan, np.nan]])

        actual = transformation.normalize(X, method=NormalizationMethod.SUM)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_single_single_row_with_all_zeros_return_all_nan(self):
        X = np.array([[0.0, 0.0, 0.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.SUM)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_single_single_row_with_all_zeros_or_nan_return_all_nan(self):
        X = np.array([[0.0, np.nan, 0.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.SUM)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_multiple_rows(self):
        X = np.array([[0.0, np.nan, 0.0], [1.0, 1.0, 1.0], [1.0, np.nan, np.nan]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.SUM)
        assert X.shape == actual.shape

        actual_row_sum = np.nansum(actual, axis=1, keepdims=True)
        expected_row_sum = np.array([[0.0], [1.0], [1.0]])
        assert np.allclose(actual_row_sum, expected_row_sum)


class TestNormalizeMax:
    def test_single_column_single_row_returns_one(self):
        X = np.array([[10.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.MAX)
        expected = np.array([[1.0]])

        assert np.allclose(actual, expected)

    def test_single_single_row(self):
        X = np.array([[1.0, 2.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.MAX)
        expected = np.array([[0.5, 1.0]])
        assert np.allclose(actual, expected)

    def test_single_single_row_with_nans(self):
        X = np.array([[1.0, 2.0, np.nan]])

        actual = transformation.normalize(X, method=NormalizationMethod.MAX)
        expected = np.array([0.5, 1.0, np.nan])
        assert np.allclose(actual, expected, equal_nan=True)

    def test_single_single_row_with_all_nans_return_all_nan(self):
        X = np.array([[np.nan, np.nan, np.nan]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.MAX)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_single_single_row_with_all_zeros_return_all_nan(self):
        X = np.array([[0.0, 0.0, 0.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.MAX)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_single_single_row_with_all_zeros_or_nan_return_all_nan(self):
        X = np.array([[0.0, np.nan, 0.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.MAX)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_multiple_rows(self):
        X = np.array([[0.0, np.nan, 0.0], [1.0, 1.0, 2.0], [1.0, np.nan, np.nan]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.MAX)

        expected = np.array([[np.nan, np.nan, np.nan], [0.5, 0.5, 1.0], [1.0, np.nan, np.nan]])
        assert np.allclose(actual, expected, equal_nan=True)


class TestNormalizeEuclidean:
    def test_single_column_single_row_returns_one(self):
        X = np.array([[10.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.EUCLIDEAN)
        expected = np.array([[1.0]])

        assert np.allclose(actual, expected)

    def test_single_single_row(self):
        X = np.array([[3.0, 4.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.EUCLIDEAN)
        expected = np.array([[0.6, 0.8]])
        assert np.allclose(actual, expected)

    def test_single_single_row_with_nans(self):
        X = np.array([[3.0, 4.0, np.nan]])

        actual = transformation.normalize(X, method=NormalizationMethod.EUCLIDEAN)
        expected = np.array([0.6, 0.8, np.nan])
        assert np.allclose(actual, expected, equal_nan=True)

    def test_single_single_row_with_all_nans_return_all_nan(self):
        X = np.array([[np.nan, np.nan, np.nan]])

        actual = transformation.normalize(X, method=NormalizationMethod.EUCLIDEAN)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_single_single_row_with_all_zeros_return_all_nan(self):
        X = np.array([[0.0, 0.0, 0.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.EUCLIDEAN)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_single_single_row_with_all_zeros_or_nan_return_all_nan(self):
        X = np.array([[0.0, np.nan, 0.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.EUCLIDEAN)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_multiple_rows(self):
        X = np.array([[0.0, np.nan, 0.0], [2.0, 3.0, 6.0], [1.0, np.nan, np.nan]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.EUCLIDEAN)

        expected = np.array([[np.nan, np.nan, np.nan], [2.0 / 7, 3.0 / 7.0, 6.0 / 7.0], [1.0, np.nan, np.nan]])
        assert np.allclose(actual, expected, equal_nan=True)


class TestNormalizeFeature:
    def test_undefined_index_raises_error(self):
        X = np.array([[10.0]])

        with pytest.raises(ValueError):
            transformation.normalize(X, method=NormalizationMethod.FEATURE)

    def test_single_column_single_row_returns_one(self):
        X = np.array([[10.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.FEATURE, index=0)
        expected = np.array([[1.0]])

        assert np.allclose(actual, expected)

    def test_single_single_row(self):
        X = np.array([[1.0, 2.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.FEATURE, index=0)
        expected = np.array([[1.0, 2.0]])
        assert np.allclose(actual, expected)

    def test_single_single_row_with_nans(self):
        X = np.array([[1.0, 2.0, np.nan]])

        actual = transformation.normalize(X, method=NormalizationMethod.FEATURE, index=0)
        expected = np.array([1.0, 2.0, np.nan])
        assert np.allclose(actual, expected, equal_nan=True)

    def test_single_single_with_feature_nan_return_all_nan(self):
        X = np.array([[np.nan, 1.0, 2.0]])

        actual = transformation.normalize(X, method=NormalizationMethod.FEATURE, index=0)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual))

    def test_single_single_row_with_feature_zeros_return_all_nan(self):
        X = np.array([[0.0, 1.0, 1.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.normalize(X, method=NormalizationMethod.FEATURE, index=0)
        assert X.shape == actual.shape
        assert np.all(np.isnan(actual[:, 0]))
        assert np.all(np.isinf(actual[:, 1:]))

    def test_multiple_rows(self):
        X = np.array([[1.0, np.nan, 0.0], [1.0, 1.0, 2.0], [1.0, np.nan, np.nan]])

        actual = transformation.normalize(X, method=NormalizationMethod.FEATURE, index=0)

        expected = np.array([[1.0, np.nan, 0.0], [1.0, 1.0, 2.0], [1.0, np.nan, np.nan]])
        assert np.allclose(actual, expected, equal_nan=True)


class TestScaleAutoscaling:
    def test_non_2d_array_raises_error(self):
        X = np.arange(5)
        with pytest.raises(ValueError):
            transformation.scale(X, method=ScalingMethod.AUTOSCALING)

    def test_2d_empty_array_raise_error(self):
        X = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            transformation.scale(X, method=ScalingMethod.AUTOSCALING)

    def test_use_str_scaling_method(self):
        X = np.array([[2.0], [0.0]])

        actual = transformation.scale(X, method="autoscaling")
        expected = np.array([[1.0], [-1.0]])

        assert np.allclose(actual, expected)

    def test_single_column_single_row_returns_nan(self):
        X = np.array([[10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.AUTOSCALING)

        assert np.all(np.isnan(actual))

    def test_constant_single_column_returns_nan(self):
        X = np.array([[10.0], [10.0], [10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.AUTOSCALING)

        assert np.all(np.isnan(actual))

    def test_single_column_nans_are_ignored_nan(self):
        X = np.array([[2.0], [np.nan], [0.0]])

        actual = transformation.scale(X, method=ScalingMethod.AUTOSCALING)
        expected = np.array([[1.0], [np.nan], [-1.0]])

        assert np.allclose(actual, expected, equal_nan=True)

    def test_single_column_no_nans(self):
        X = np.array([[2.0], [0.0]])

        actual = transformation.scale(X, method=ScalingMethod.AUTOSCALING)
        expected = np.array([[1.0], [-1.0]])

        assert np.allclose(actual, expected)

    def test_multiple_columns_ok(self):
        X = np.array([[2.0, 10.0], [0.0, 10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.AUTOSCALING)
        expected = np.array([[1.0, np.nan], [-1.0, np.nan]])

        assert np.allclose(actual, expected, equal_nan=True)


class TestScaleRescaling:
    def test_non_2d_array_raises_error(self):
        X = np.arange(5)
        with pytest.raises(ValueError):
            transformation.scale(X, method=ScalingMethod.RESCALING)

    def test_2d_empty_array_raise_error(self):
        X = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            transformation.scale(X, method=ScalingMethod.RESCALING)

    def test_single_column_single_row_returns_nan(self):
        X = np.array([[10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.RESCALING)

        assert np.all(np.isnan(actual))

    def test_constant_single_column_returns_nan(self):
        X = np.array([[10.0], [10.0], [10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.AUTOSCALING)

        assert np.all(np.isnan(actual))

    def test_single_column_nans_are_ignored_nan(self):
        X = np.array([[2.0], [np.nan], [0.0]])

        actual = transformation.scale(X, method=ScalingMethod.RESCALING)
        expected = np.array([[1.0], [np.nan], [0.0]])

        assert np.allclose(actual, expected, equal_nan=True)

    def test_single_column_no_nans(self):
        X = np.array([[2.0], [0.0]])

        actual = transformation.scale(X, method=ScalingMethod.RESCALING)
        expected = np.array([[1.0], [0.0]])

        assert np.allclose(actual, expected)

    def test_multiple_columns_ok(self):
        X = np.array([[2.0, 10.0], [0.0, 10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.RESCALING)
        expected = np.array([[1.0, np.nan], [0.0, np.nan]])

        assert np.allclose(actual, expected, equal_nan=True)


class TestScalePareto:
    def test_non_2d_array_raises_error(self):
        X = np.arange(5)
        with pytest.raises(ValueError):
            transformation.scale(X, method=ScalingMethod.PARETO)

    def test_2d_empty_array_raise_error(self):
        X = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            transformation.scale(X, method=ScalingMethod.PARETO)

    def test_single_column_single_row_returns_nan(self):
        X = np.array([[10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.PARETO)

        assert np.all(np.isnan(actual))

    def test_constant_single_column_returns_nan(self):
        X = np.array([[10.0], [10.0], [10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.PARETO)

        assert np.all(np.isnan(actual))

    def test_single_column_nans_are_ignored_nan(self):
        X = np.array([[2.0], [np.nan], [0.0]])

        actual = transformation.scale(X, method=ScalingMethod.PARETO)
        expected = np.array([[1.0], [np.nan], [-1.0]])

        assert np.allclose(actual, expected, equal_nan=True)

    def test_single_column_no_nans(self):
        X = np.array([[2.0], [0.0]])

        actual = transformation.scale(X, method=ScalingMethod.PARETO)
        expected = np.array([[1.0], [-1.0]])

        assert np.allclose(actual, expected)

    def test_multiple_columns_ok(self):
        X = np.array([[2.0, 10.0], [0.0, 10.0]])

        with pytest.warns(RuntimeWarning):
            actual = transformation.scale(X, method=ScalingMethod.PARETO)
        expected = np.array([[1.0, np.nan], [-1.0, np.nan]])

        assert np.allclose(actual, expected, equal_nan=True)



"""Download example datasets."""

import pathlib
from typing import Any

import httpx
import pydantic

from ..core.exceptions import DatasetNotFound


class DatasetFileContent(pydantic.BaseModel):
    """Store a dataset file content."""

    name: str
    """The file name"""

    path: str
    """The file path in the GitHub repo"""

    content: str | None = None


def download_dataset(name: str, download_dir: pathlib.Path | str, token: str | None = None):
    """Download a list of files from the `data repository <https://github.com/griquelme/tidyms-data>`_.

    :param name: name of the data directory
    :param download_dir : path to download the data.
    :param token: A GitHub personal access token

    Examples
    --------
    Download the `data.csv` file from the `reference-materials` directory into
    the current directory:

    >>> import tidyms as ms
    >>> dataset = "reference-materials"
    >>> ms.fileio.download_tidyms_data(dataset, file_list, download_dir=".")

    See Also
    --------
    download_dataset
    load_dataset

    """
    owner = "griquelme"
    repo = "tidyms-data"

    dataset_files = list_dataset_files(owner, repo, name, token=token)

    if isinstance(download_dir, str):
        download_dir = pathlib.Path(download_dir)

    download_dir.mkdir(exist_ok=True, parents=True)

    for file_content in dataset_files:
        file_path = download_dir / file_content.name
        if file_path.is_file():
            continue

        if file_content.content is None:
            file_content.content = _get_github_repo_contents(owner, repo, file_content.path, raw=True, token=token)

        file_path.write_text(file_content.content, newline="")  # type: ignore


def get_dataset_file(dataset: str, file: str, download_path: pathlib.Path) -> pathlib.Path:
    """Get the path to an example dataset file."""
    download_dataset(dataset, download_path)

    path = download_path / dataset / file

    if not path.is_file():
        msg = f"File {file} not found in dataset {dataset}."
        raise DatasetNotFound(msg)

    return path


def list_dataset_files(owner: str, repo: str, dataset: str, token: str | None = None) -> list[DatasetFileContent]:
    """List the files in a directory in a GitHub repository."""
    try:
        contents = _get_github_repo_contents(owner, repo, dataset, raw=False, token=token)
    except (httpx.HTTPError, ValueError):
        raise DatasetNotFound(dataset)

    if not isinstance(contents, list):
        raise DatasetNotFound(dataset)

    return [DatasetFileContent(**x) for x in contents]


def _get_github_repo_contents(owner: str, repo: str, path: str, raw: bool = False, token: str | None = None) -> Any:
    endpoint = f"https://api.github.com/repos/{owner}/{repo}/contents/{path}"
    headers = {"Accept": "application/vnd.github.raw+json"} if raw else dict()
    if token is not None:
        headers["Authorization"] = f"Bearer {token}"

    response = httpx.get(endpoint, headers=headers)

    if not response.is_success:
        response.raise_for_status()
    return response.text if raw else response.json()



"""Functions to find isotopic envelopes candidates in a list of m/z values."""

import bisect
from collections.abc import Sequence

from ..chem import ChemicalContext
from ..chem.atoms import Element
from ..core.models import AnnotableFeatureType, Feature
from .annotation_data import AnnotationData
from .config import AnnotatorParameters

# name conventions
# M is used for Molecular mass
# m for nominal mass
# p for abundances


class EnvelopeFinder:
    r"""Find isotopic envelopes candidates in a list of sorted m/z values.

    :param elements : the list of elements used to compute mass difference windows.
    :param mz_tolerance : m/z tolerance used to match candidates.
    :param max_length: maximum envelope length to search.
    :param min_p: the minimum abundance of the isotopes of each element to be used
        for m/z estimation. Must be a number between 0 and 1.
    min_similarity : minimum similarity to create candidates.

    Notes
    -----
    Using a list of elements, theoretical bounds are computed for each M1, M2,
    M3, etc... isotopologue. Then using these values and the `mz_tolerance` and
    the `max_charge`, the bounds are adjusted according to the following
    equations:

    .. math::

        mz_{k, min}= \frac{m_{k, min}{q} - mz_{tolerance}

        mz_{k, max}= \frac{m_{k, max}{q} + mz_{tolerance}

    where :math:`m_{k, min}` is the minimum theoretical value for the k-th
    isotopologue and q is the charge.

    The envelopes candidates found are determined based on m/z compatibility
    only. To reduce the number of candidates, the list of m/z values should be
    reduced by other means, such as correlation of the values.

    """

    def __init__(self, config: AnnotatorParameters, context: ChemicalContext):
        self.config = config
        elements = [context.table.get_element(x) for x in config.bounds]
        self.bounds = _make_exact_mass_difference_bounds(elements, config.min_p)

    def find(
        self, data: AnnotationData[AnnotableFeatureType], mmi: AnnotableFeatureType, charge: int
    ) -> list[Sequence[AnnotableFeatureType]]:
        """Find isotopic envelope candidates starting from the minimum mass isotopologue (MMI).

        :param data: annotation data with features used in the search
        :param mmi: a feature in `data used as the :term:`mmi`.
        :param charge: the absolute value of the charge state of the isotopic envelope

        """
        completed_candidates = list()

        candidates = [[mmi]]
        while candidates:
            # remove and extend a candidate
            candidate = candidates.pop()

            # find features with compatible m/z and similarities
            min_mz, max_mz = _get_next_mz_search_interval(candidate, self.bounds, charge, self.config.max_M_tol)
            start = bisect.bisect(data.features, min_mz, key=lambda x: x.mz)
            end = bisect.bisect(data.features, max_mz, key=lambda x: x.mz)
            new_features = list()
            for k in range(start, end):
                k_ft = data.features[k]
                is_similar = data.similarity_cache.get_similarity(mmi, k_ft) >= self.config.min_similarity
                is_non_annotated = k_ft in data.non_annotated
                if is_similar and is_non_annotated:
                    new_features.append(k_ft)

            # extend candidates with compatible features
            length = len(candidate)
            if new_features and (length < self.config.max_length):
                tmp = [candidate + [x] for x in new_features]
                candidates.extend(tmp)
            else:
                completed_candidates.append(candidate)
        completed_candidates = [x for x in completed_candidates if len(x) > 1]
        return _remove_sub_candidates(completed_candidates)


def _remove_sub_candidates(candidates: list[Sequence[AnnotableFeatureType]]) -> list[Sequence[AnnotableFeatureType]]:
    """Remove candidates that are subsets of other candidates."""
    validated = list()
    while candidates:
        last = candidates.pop()
        last_set = set(last)
        is_subset = False
        for candidate in candidates:
            is_subset = last_set <= set(candidate)
        if not is_subset:
            validated.append(last)
    return validated


def _get_next_mz_search_interval(
    envelope: Sequence[Feature],
    elements_mass_difference: dict[int, tuple[float, float]],
    charge: int,
    mz_tolerance: float,
) -> tuple[float, float]:
    """Compute the valid m/z range for a k-th isotopologue using previous isotopologues information.

    :return: a tuple with the minimum and maximum m/z values expected for the next isotopologue in
        the envelope.

    """
    # If the charge is 0 (neutral mass) the results are the same as using
    # charge = 1. There is no difference between positive and negative
    # charges
    charge = max(1, abs(charge))
    length = len(envelope)
    min_mz = envelope[-1].mz + 2  # dummy values
    max_mz = envelope[-1].mz - 2
    for dm, (min_dM, max_dM) in elements_mass_difference.items():
        i = length - dm
        if i >= 0:
            min_mz = min(min_mz, envelope[i].mz + min_dM / charge)
            max_mz = max(max_mz, envelope[i].mz + max_dM / charge)
    min_mz -= mz_tolerance
    max_mz += mz_tolerance
    return min_mz, max_mz


def _make_exact_mass_difference_bounds(elements: list[Element], min_p: float) -> dict[int, tuple[float, float]]:
    """Compute possible mass differences obtaining from changing one isotope."""
    dm_to_dM_list: dict[int, list[float]] = dict()
    for e in elements:
        for isotope in e.isotopes:
            if isotope == e.mmi or isotope.p < min_p:
                continue
            dm = isotope.a - e.mmi.a
            dM = isotope.m - e.mmi.m
            dM_list = dm_to_dM_list.setdefault(dm, list())
            dM_list.append(dM)

    return {k: (min(v), max(v)) for k, v in dm_to_dM_list.items()}



"""Storage interface for sample data and assay data."""

from __future__ import annotations

from itertools import product
from typing import Generic, Iterable, Protocol, TypeVar
from uuid import UUID

from .dataflow import AssayProcessStatus, SampleProcessStatus
from .models import Annotation, AnnotationPatch, DescriptorPatch, FeatureGroup, FeatureType, FillValue, RoiType, Sample


class SampleStorage(Protocol, Generic[RoiType, FeatureType]):
    """Interface to store sample data.

    Provides:

    Access sample metadata.
        Sample metadata maintains all sample-related information. It also provides utilities
        to access raw data.
    add/get/delete ROIs and features.
        Provide utilities to modify the sample data.
    sample data status
        `get_status` retrieves a record of the information available for the sample,
        such as if ROI extraction was performed on the data, feature extraction,
        isotopologue annotation, etc... `set_status` allow to specify this status.
    linear data snapshots
        Sample storage maintains a list of snapshots with independent copies of the data
        at different points in time. The `create_snapshot` allows to create data snapshots
        of the sample data at any point in time. Only the latest snapshot is modifiable.
        The `reset` method rollbacks the data to a specified snapshot in read only mode,
        i.e., all operations that modify the snapshot state should raise SnapshotError.
        The ``hard`` parameter controls if a hard reset operation is performed on the data.
        That is, setting the selected snapshot as the latest and deleting data from all
        posterior snapshots. Finally, the `set_latest_snapshot` set the data to the latest
        snapshot.

    """

    roi_type: type[RoiType]
    feature_type: type[FeatureType]

    def __init__(self, sample: Sample, roi_type: type[RoiType], feature_type: type[FeatureType], **kwargs) -> None: ...

    def add_features(self, *features: FeatureType) -> None:
        """Add features to the sample storage."""
        ...

    def add_rois(self, *rois: RoiType) -> None:
        """Add ROIs to the sample storage."""
        ...

    def create_snapshot(self, snapshot_id: str) -> None:
        """Create a snapshot of the sample data."""
        ...

    def delete_features(self, *feature_ids: UUID) -> None:
        """Delete features from the sample storage using their ids."""
        ...

    def delete_rois(self, *roi_id: UUID) -> None:
        """Delete ROIs from the sample storage using their ids."""
        ...

    def get_feature(self, feature_id: UUID) -> FeatureType:
        """Get a feature from the sample storage using its id."""
        ...

    def get_n_features(self) -> int:
        """Get the total number of features in the storage."""
        ...

    def get_n_rois(self) -> int:
        """Get the total number of ROIs in the storage."""
        ...

    def get_roi(self, roi_id: UUID) -> RoiType:
        """Get a ROI from the sample storage using its id."""
        ...

    def get_sample(self) -> Sample:
        """Get the sample metadata."""
        ...

    def get_snapshot_id(self) -> str:
        """Get the current snapshot id."""
        ...

    def get_status(self) -> SampleProcessStatus:
        """Get the current process status."""
        ...

    def set_status(self, status: SampleProcessStatus) -> None:
        """Set the current process status."""
        ...

    def has_feature(self, feature_id: UUID) -> bool:
        """Check if a feature with the provided id exists."""
        ...

    def has_roi(self, roi_id: UUID) -> bool:
        """Check if a ROI with the provided id exists."""
        ...

    def list_features(self, roi_id: UUID | None = None) -> list[FeatureType]:
        """List features extracted from the sample."""
        ...

    def list_rois(self) -> list[RoiType]:
        """List ROIs extracted from the sample."""
        ...

    def list_snapshots(self) -> list[str]:
        """List all available snapshots."""
        ...

    def set_snapshot(self, snapshot_id: str | None = None, reset: bool = False) -> None:
        """Set the sample data to the specified snapshot id.

        Setting `reset` to ``True`` set the selected status as latest and deletes all posterior snapshots.

        """
        ...


class AssayStorage(Protocol, Generic[RoiType, FeatureType]):
    """Interface to store assay data."""

    def add_feature_groups(self, *groups: FeatureGroup) -> None:
        """Add feature groups data to the assay storage."""
        ...

    def add_fill_values(self, *fill_values: FillValue) -> None:
        """Add values to fill missing entries in the data matrix."""
        ...

    def add_sample_data(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        """Add data from a sample.

        If include snapshots is set to ``True`` include all data snapshots. Otherwise, include only the
        latest data snapshot.

        """
        ...

    def create_snapshot(self, snapshot_id: str) -> None:
        """Create a new assay data snapshot."""
        ...

    def fetch_annotations(self, sample_id: str | None = None) -> list[Annotation]:
        """Fetch the feature annotations."""
        ...

    def fetch_descriptors(self, sample_id: str | None = None, descriptors: Iterable[str] | None = None) -> dict:
        """Fetch the feature descriptors."""
        ...

    def fetch_features_by_group(self, group: int) -> list[FeatureType]:
        """Fetch features using the feature group id."""
        ...

    def fetch_features_by_id(self, *feature_ids: UUID) -> list[FeatureType]:
        """Fetch features using their ids."""
        ...

    def fetch_features_by_sample(self, sample_id: str) -> list[FeatureType]:
        """Fetch features using the sample id."""
        ...

    def fetch_feature_groups(self) -> list[FeatureGroup]:
        """Fetch feature groups stored in the assay."""
        ...

    def fetch_fill_values(self) -> dict[str, dict[int, float]]:
        """Fetch fill values for data matrix."""
        ...

    def fetch_rois_by_id(self, *roi_ids: UUID) -> list[RoiType]:
        """Fetch ROIs using their ids."""
        ...

    def fetch_rois_by_sample(self, sample_id: str) -> list[RoiType]:
        """Fetch all ROIs from a sample."""
        ...

    def fetch_sample(self, sample_id: str) -> Sample:
        """Fetch sample metadata from a sample using its id."""
        ...

    def fetch_sample_data(self, sample_id: str) -> SampleStorage[RoiType, FeatureType]:
        """Fetch Samples from the assay using their ids."""
        ...

    def get_n_features(self) -> int:
        """Get the total number of features in the assay."""
        ...

    def get_n_rois(self) -> int:
        """Get the total number of ROIs in the assay."""
        ...

    def get_process_status(self) -> AssayProcessStatus:
        """Get the current process status."""
        ...

    def get_snapshot_id(self) -> str:
        """Retrieve the current snapshot id."""
        ...

    def get_feature_type(self) -> type[FeatureType]:
        """Retrieve the Feature class used."""
        ...

    def get_roi_type(self) -> type[RoiType]:
        """Retrieve the ROI class used."""
        ...

    def has_feature(self, feature_id: UUID) -> bool:
        """Check if a feature with the provided id exists."""
        ...

    def has_feature_group(self, feature_group: int) -> bool:
        """Check if a group with the provided id is in the assay."""
        ...

    def has_roi(self, roi_id: UUID) -> bool:
        """Check if a ROI with the provided id exists."""
        ...

    def has_sample(self, sample_id: str) -> bool:
        """Check if a sample with the provided id is in the assay."""
        ...

    def list_feature_groups(self) -> list[int]:
        """List all group ids in the assay."""
        ...

    def list_samples(self) -> list[Sample]:
        """Fetch metadata from all samples in the assay."""
        ...

    def list_snapshots(self) -> list[str]:
        """Retrieve the list of all snapshots."""
        ...

    def patch_annotations(self, *patches: AnnotationPatch) -> None:
        """Update feature annotation values."""
        ...

    def patch_descriptors(self, *patches: DescriptorPatch) -> None:
        """Update feature descriptors values."""
        ...

    def set_process_status(self, status: AssayProcessStatus) -> None:
        """Set the new process status."""
        ...

    def set_snapshot(self, snapshot_id: str | None = None, reset: bool = False) -> None:
        """Set assay storage data to specified snapshot.

        If ``None``, fetch data from the latest snapshot.

        """
        ...


def find_missing_entries(assay_storage: AssayStorage) -> dict[str, list[int]]:
    """List all data matrix missing entries from data in an assay.

    Entries are sorted by sample_id and feature group.

    """
    sample_ids = [x.id for x in assay_storage.list_samples()]
    group_ids = assay_storage.list_feature_groups()
    all_entries = set(product(sample_ids, group_ids))
    existing_entries = {(x.sample_id, x.group) for x in assay_storage.fetch_annotations() if x.group > -1}
    missing_entries = dict()

    for sample, group in all_entries.difference(existing_entries):
        sample_list = missing_entries.setdefault(sample, list())
        sample_list.append(group)
    return missing_entries


def delete_empty_rois(sample_storage: SampleStorage[RoiType, FeatureType]) -> None:
    """Remove ROI that do not contain any features from the sample storage.

    :param sample_storage: any instance that implements the SampleStorage protocol.

    """
    roi_with_feature_ids = {x.roi.id for x in sample_storage.list_features()}
    all_roi_ids = {x.id for x in sample_storage.list_rois()}
    empty_roi_ids = all_roi_ids.difference(roi_with_feature_ids)
    sample_storage.delete_rois(*empty_roi_ids)


_SAMPLE_STORAGES: dict[str, SampleStorage] = dict()


T = TypeVar("T")


def register_sample_storage(storage_type: type[T]) -> type[T]:
    """Register a sample storage.

    :param storage_type: A class implementing the SampleStorage protocol to register

    """
    if storage_type.__name__ in _SAMPLE_STORAGES:
        raise ValueError
    return storage_type



import pathlib

import numpy as np
import pytest

from tidyms2.algorithms import raw
from tidyms2.core.models import Chromatogram, MSSpectrum, Sample
from tidyms2.io import MSData


@pytest.fixture
def sample(lcms_sample_factory) -> Sample:
    return lcms_sample_factory("sample")


@pytest.fixture
def ms_data(sample: Sample):
    return MSData(sample)


@pytest.fixture
def mz_list(lcms_sample_factory) -> list[float]:
    sample_spec = lcms_sample_factory.create_simulated_sample_spec("default", 0)
    return list(sample_spec.make_grid())


class TestMakeRoi:
    @pytest.fixture
    def params(self) -> raw.MakeRoiParameters:
        return raw.MakeRoiParameters(tolerance=0.01, max_missing=0, min_length=5)

    def test_ok(self, ms_data: MSData, mz_list: list[float], params):
        roi_list = raw.make_roi(ms_data, params)
        assert len(roi_list) == len(mz_list)

    def test_targeted_mz_ok(self, ms_data: MSData, mz_list: list[float], params):
        # the first three m/z values generated by simulated experiment are used
        params.targeted_mz = mz_list
        params.targeted_mz.pop()
        roi_list = raw.make_roi(ms_data, params)
        assert len(roi_list) == len(params.targeted_mz)

    def test_make_roi_min_intensity(self, ms_data, params):
        # only two features have intensity larger than 2000
        expected_n_roi = 1
        params.min_intensity = 2000.0
        roi_list = raw.make_roi(ms_data, params)
        assert len(roi_list) == expected_n_roi

    def test_make_roi_multiple_match_closest(self, ms_data: MSData, params, mz_list):
        params.multiple_match = "closest"
        params.tolerance = 2.0  # Make match conflict between isotopologues
        roi_list = raw.make_roi(ms_data, params)
        assert len(roi_list) == len(mz_list)

    def test_make_roi_multiple_match_reduce(self, ms_data: MSData, params, mz_list):
        params.multiple_match = "reduce"
        params.targeted_mz = mz_list[::2]
        params.tolerance = 2.0  # Make match conflict between isotopologues
        roi_list = raw.make_roi(ms_data, params)
        assert len(roi_list) == len(mz_list) // 2


class TestAccumulateSpectra:
    def test_centroid(self, ms_data, mz_list):
        params = raw.AccumulateSpectraParameters(start_time=0.0, end_time=50.0)
        cum_sp = raw.accumulate_spectra(ms_data, params)
        assert cum_sp.mz.size == len(mz_list)


class TestMakeChromatograms:
    def test_make_chromatograms_using_sum_accumulator(self, ms_data: MSData, mz_list):
        # test that the chromatograms generated are valid

        # create chromatograms
        n_sp = ms_data.get_n_spectra()

        params = raw.MakeChromatogramParameters(mz=mz_list)

        chromatograms = raw.make_chromatograms(ms_data, params)
        assert len(chromatograms) == len(params.mz)
        for chrom in chromatograms:
            assert np.all(np.diff(chrom.time) > 0.0)
            assert chrom.time.size == n_sp

    def test_empty_mz_list_returns_empty_list(self, ms_data: MSData):
        params = raw.MakeChromatogramParameters(mz=list())
        actual = raw.make_chromatograms(ms_data, params)
        assert isinstance(actual, list)
        assert not actual

    def test_mz_list_without_signals_returns_all_zeroes(self, ms_data: MSData):
        params = raw.MakeChromatogramParameters(mz=[2500.0])

        chromatogram = raw.make_chromatograms(ms_data, params)[0]
        assert np.allclose(chromatogram.int, 0.0)

    def test_mz_list_without_signals_and_no_fill_missing_returns_with_all_nan(self, ms_data: MSData):
        params = raw.MakeChromatogramParameters(mz=[2500.0], fill_missing=False)
        chromatogram = raw.make_chromatograms(ms_data, params)[0]
        assert np.all(np.isnan(chromatogram.int))


class TestMakeTIC:
    def test_tic(self, ms_data: MSData):
        params = raw.MakeTICParameters(kind="tic")
        tic = raw.make_tic(ms_data, params)
        assert isinstance(tic, Chromatogram)
        assert np.all(tic.int >= 0.0)

    def test_bpi(self, ms_data: MSData):
        params = raw.MakeTICParameters(kind="bpi")
        bpi = raw.make_tic(ms_data, params)
        assert isinstance(bpi, Chromatogram)
        assert np.all(bpi.int >= 0.0)

    def test_bpi_intensity_is_lower_or_equal_than_tic_intensity(self, ms_data: MSData):
        bpi = raw.make_tic(ms_data, raw.MakeTICParameters(kind="bpi"))
        tic = raw.make_tic(ms_data, raw.MakeTICParameters(kind="tic"))
        assert np.all(tic.int >= bpi.int)


class TestTempRoi:
    def test_creation(self):
        temp_roi = raw._TempRoi()
        assert len(temp_roi.mz) == 0
        assert len(temp_roi.spint) == 0
        assert len(temp_roi.scan) == 0

    def test_append_ok(self):
        mz, sp, scan = 150.0, 200.0, 10
        temp_roi = raw._TempRoi()
        temp_roi.append(mz, sp, scan)

        assert temp_roi.mz[-1] == mz
        assert temp_roi.spint[-1] == sp
        assert temp_roi.scan[-1] == scan

    def test_clear_empties_temp_roi(self):
        mz, sp, scan = 150.0, 200.0, 10
        temp_roi = raw._TempRoi()
        temp_roi.append(mz, sp, scan)
        temp_roi.clear()

        assert len(temp_roi.mz) == 0
        assert len(temp_roi.spint) == 0
        assert len(temp_roi.scan) == 0

    def test_pad_appends_nan_on_both_ends(self):
        temp_roi = raw._TempRoi()
        temp_roi.append(1, 1, 2)
        temp_roi.append(1, 1, 3)
        scans = np.array([0, 1, 2, 3, 4, 5])
        temp_roi.pad(2, scans)
        assert list(temp_roi.mz) == [np.nan, np.nan, 1, 1, np.nan, np.nan]
        assert list(temp_roi.spint) == [np.nan, np.nan, 1, 1, np.nan, np.nan]
        assert list(temp_roi.scan) == list(scans)

    def test_pad_stops_on_left_scan_boundary(self):
        temp_roi = raw._TempRoi()
        temp_roi.append(1, 1, 1)
        temp_roi.append(1, 1, 2)
        scans = np.array([0, 1, 2, 3, 4, 5])
        temp_roi.pad(2, scans)
        # 0 is the minimum scan number, only one element should be padded to the left
        assert list(temp_roi.mz) == [np.nan, 1, 1, np.nan, np.nan]
        assert list(temp_roi.spint) == [np.nan, 1, 1, np.nan, np.nan]
        assert list(temp_roi.scan) == list([0, 1, 2, 3, 4])

    def test_pad_no_left_pad_if_roi_starts_on_boundary(self):
        temp_roi = raw._TempRoi()
        temp_roi.append(1, 1, 0)
        temp_roi.append(1, 1, 1)
        scans = np.arange(6)
        temp_roi.pad(2, scans)
        # 0 is the minimum scan number, no elements should be padded to the left
        assert list(temp_roi.mz) == [1, 1, np.nan, np.nan]
        assert list(temp_roi.spint) == [1, 1, np.nan, np.nan]
        assert list(temp_roi.scan) == list([0, 1, 2, 3])

    def test_pad_stops_on_right_scan_boundary(self):
        temp_roi = raw._TempRoi()
        temp_roi.append(1, 1, 3)
        temp_roi.append(1, 1, 4)
        scans = np.arange(6)
        temp_roi.pad(2, scans)
        # 5 is the maximum scan number, right pad should add only one element
        assert list(temp_roi.mz) == [np.nan, np.nan, 1, 1, np.nan]
        assert list(temp_roi.spint) == [np.nan, np.nan, 1, 1, np.nan]
        assert list(temp_roi.scan) == list([1, 2, 3, 4, 5])

    def test_pad_no_right_pad_if_roi_ends_on_boundary(self):
        temp_roi = raw._TempRoi()
        temp_roi.append(1, 1, 4)
        temp_roi.append(1, 1, 5)
        scans = np.arange(5)
        temp_roi.pad(2, scans)
        # 5 is the maximum scan number, no elements should be padded to the right
        assert list(temp_roi.mz) == [np.nan, np.nan, 1, 1]
        assert list(temp_roi.spint) == [np.nan, np.nan, 1, 1]
        assert list(temp_roi.scan) == list([2, 3, 4, 5])

    def test_convert_to_mz_trace(self):
        temp_roi = raw._TempRoi()
        temp_roi.append(1, 1, 2)
        temp_roi.append(1, 1, 4)
        temp_roi.append(1, 1, 5)
        scans = np.arange(6)
        time = np.arange(6)
        sample = Sample(id="test", path=pathlib.Path("."))
        mz_trace = temp_roi.convert_to_roi(time, scans, sample)

        assert np.array_equal(mz_trace.scan, [2, 3, 4, 5])
        assert np.allclose(mz_trace.time, [2, 3, 4, 5])


class TestRoiList:
    def test_creation(self):
        raw._TempRoiList()
        assert True

    def test_insert_empty_RoiList(self):
        roi_list = raw._TempRoiList()
        n = 10
        sp = MSSpectrum(mz=np.arange(n), int=np.arange(n), index=1)
        roi_list.insert(sp)

        # check tracking values
        assert np.array_equal(sp.mz, roi_list.mz_mean)
        assert np.array_equal(sp.mz, roi_list.mz_sum)
        assert np.array_equal(sp.int, roi_list.max_int)
        assert (roi_list.length == 1).all()
        assert (roi_list.missing_count == 0).all()

        for r, r_mz, r_sp in zip(roi_list.roi, sp.mz, sp.int):
            assert len(r.mz) == 1
            assert len(r.spint) == 1
            assert len(r.scan) == 1
            assert r.mz[-1] == r_mz
            assert r.spint[-1] == r_sp
            assert r.scan[-1] == sp.index

    def test_RoiList_two_consecutive_insert(self):
        roi_list = raw._TempRoiList()
        # first insert
        mz1 = np.array([0, 1, 2, 3, 4, 6, 7, 9])  # 5 and 8 are missing
        sp1 = MSSpectrum(mz=mz1, int=mz1.copy(), index=1)
        roi_list.insert(sp1)

        # second insert
        mz2 = np.array([5, 8])
        sp2 = MSSpectrum(mz=mz2, int=mz2.copy(), index=2)
        roi_list.insert(sp2)

        # check tracking values
        expected_mz = np.arange(10)
        expected_missing_count = np.zeros_like(roi_list.missing_count)
        expected_missing_count[mz1] = 1
        assert np.allclose(roi_list.mz_mean, expected_mz)
        assert np.allclose(roi_list.mz_sum, expected_mz)
        assert np.allclose(roi_list.max_int, expected_mz)
        assert (roi_list.length == 1).all()
        assert (roi_list.missing_count == 0).all()
        assert (np.diff(roi_list.mz_mean) >= 0).all()  # check is sorted

        # check roi values
        # values for mz and spint should be 0, 1, 2, 3, ...
        for expected, r in enumerate(roi_list.roi):
            assert r.mz[-1] == expected
            assert r.spint[-1] == expected
            if expected in [5, 8]:
                assert r.scan[-1] == sp2.index
            else:
                assert r.scan[-1] == sp1.index
            assert len(r.mz) == 1
            assert len(r.spint) == 1
            assert len(r.scan) == 1


def assert_match(actual: raw.SpectrumMatch, expected: raw.SpectrumMatch):
    """Test equality of match objects."""
    assert np.array_equal(expected.index, actual.index)
    assert np.array_equal(expected.match.mz, actual.match.mz)
    assert np.array_equal(expected.match.int, actual.match.int)
    assert np.array_equal(expected.no_match.mz, actual.no_match.mz)
    assert np.array_equal(expected.no_match.int, actual.no_match.int)


class TestMatchMz:
    tolerance = 2

    @pytest.fixture(scope="class")
    def mz1(self):
        return np.array([50, 75, 100, 125, 150], dtype=float)

    def test_no_multiple_matches(self, mz1):
        mz2 = np.array([40, 51, 78, 91, 99, 130, 150])
        sp2 = MSSpectrum(mz=mz2, int=np.ones_like(mz2))
        mode = "closest"
        actual = raw._match_mz(mz1, sp2, self.tolerance, mode)

        mz1_match_index = np.array([0, 2, 4], dtype=int)
        mz2_match_index = np.array([1, 4, 6], dtype=int)
        mz2_no_match_index = np.array([0, 2, 3, 5], dtype=int)
        expected = raw.SpectrumMatch(
            index=mz1_match_index,
            match=MSSpectrum(mz=sp2.mz[mz2_match_index], int=sp2.int[mz2_match_index]),
            no_match=MSSpectrum(mz=sp2.mz[mz2_no_match_index], int=sp2.int[mz2_no_match_index]),
        )
        assert_match(actual, expected)

    def test_no_matches(self, mz1):
        mz2 = np.array([40, 53, 78, 91, 97, 130, 154])
        sp2 = MSSpectrum(mz=mz2, int=np.ones_like(mz2))

        # expected values for match/no match indices
        mz1_match_index = np.array([], dtype=int)
        mz2_match_index = np.array([], dtype=int)
        mz2_no_match_index = np.array([0, 1, 2, 3, 4, 5, 6], dtype=int)
        expected = raw.SpectrumMatch(
            index=mz1_match_index,
            match=MSSpectrum(mz=sp2.mz[mz2_match_index], int=sp2.int[mz2_match_index]),
            no_match=MSSpectrum(mz=sp2.mz[mz2_no_match_index], int=sp2.int[mz2_no_match_index]),
        )

        mode = "closest"
        actual = raw._match_mz(mz1, sp2, self.tolerance, mode)

        assert_match(actual, expected)

    def test_all_match(self, mz1):
        mz2 = np.array([51, 77, 99, 126, 150])
        sp2 = MSSpectrum(mz=mz2, int=np.ones_like(mz2))

        mz1_match_index = np.array([0, 1, 2, 3, 4], dtype=int)
        mz2_match_index = np.array([0, 1, 2, 3, 4], dtype=int)
        mz2_no_match_index = np.array([], dtype=int)
        expected = raw.SpectrumMatch(
            index=mz1_match_index,
            match=MSSpectrum(mz=sp2.mz[mz2_match_index], int=sp2.int[mz2_match_index]),
            no_match=MSSpectrum(mz=sp2.mz[mz2_no_match_index], int=sp2.int[mz2_no_match_index]),
        )

        mode = "closest"
        actual = raw._match_mz(mz1, sp2, self.tolerance, mode)

        assert_match(actual, expected)

    def test_match_mz_multiple_matches_mode_closest(self, mz1):
        mz2 = np.array([49, 51, 78, 99, 100, 101, 126, 150, 151])
        sp2 = MSSpectrum(mz=mz2, int=np.ones_like(mz2))
        # expected values for match/no match indices
        # in `closest` mode, argmin is used to select the closest value. If more
        # than one value has the same difference, the first one in the array is
        # going to be selected.
        mz1_match_index = np.array([0, 2, 3, 4], dtype=int)
        mz2_match_index = np.array([0, 4, 6, 7], dtype=int)
        mz2_no_match_index = np.array([1, 2, 3, 5, 8], dtype=int)
        expected = raw.SpectrumMatch(
            index=mz1_match_index,
            match=MSSpectrum(mz=sp2.mz[mz2_match_index], int=sp2.int[mz2_match_index]),
            no_match=MSSpectrum(mz=sp2.mz[mz2_no_match_index], int=sp2.int[mz2_no_match_index]),
        )

        mode = "closest"
        actual = raw._match_mz(mz1, sp2, self.tolerance, mode)

        assert_match(actual, expected)

    def test_match_mz_multiple_matches_mode_reduce(self, mz1):
        mz2 = np.array([49, 51, 78, 99, 100, 101, 126, 150, 151], dtype=float)
        sp2 = MSSpectrum(mz=mz2, int=np.ones_like(mz2))
        # expected values for match/no match indices
        # in `closest` mode, argmin is used to select the closest value. If more
        # than one value has the same difference, the first one in the array is
        # going to be selected.
        mz1_match_index = np.array([0, 2, 3, 4], dtype=int)
        mz2_no_match_index = np.array([2], dtype=int)
        expected_mz2_match = np.array([50.0, 100.0, 126.0, 150.5])
        expected_sp2_match = np.array([2, 3, 1, 2])
        expected = raw.SpectrumMatch(
            index=mz1_match_index,
            match=MSSpectrum(mz=expected_mz2_match, int=expected_sp2_match),
            no_match=MSSpectrum(mz=sp2.mz[mz2_no_match_index], int=sp2.int[mz2_no_match_index]),
        )

        mode = "reduce"
        actual = raw._match_mz(mz1, sp2, self.tolerance, mode)

        assert_match(actual, expected)

    def test_match_mz_invalid_mode(self, mz1):
        mz2 = np.array([49, 51, 78, 99, 100, 101, 126, 150, 151])
        sp2 = MSSpectrum(mz=mz2, int=np.ones_like(mz2))
        # expected values for match/no match indices
        # in closest mode, argmin is used to select the closest value. If more
        # than one value has the same difference, the first one in the array is
        # going to be selected.
        mode = "invalid-mode"
        with pytest.raises(ValueError):
            raw._match_mz(mz1, sp2, self.tolerance, mode)


class TestComputeSliceVariance:
    size = 100

    @pytest.fixture
    def x(self):
        return np.random.normal(size=self.size)

    @pytest.mark.parametrize("index", [0, 50, size])
    def test_compute_slice_variance_on_empty_slice_returns_zero(self, x, index):
        start = index
        end = index

        expected_mean = 0.0
        expected_var = 0.0

        x_cumsum = np.cumsum(x)
        x2_cumsum = np.cumsum(x**2)
        actual_mean, actual_var = raw._compute_slice_stats(x_cumsum, x2_cumsum, start, end)

        assert np.isclose(expected_mean, actual_mean)
        assert np.isclose(expected_var, actual_var)

    def test_compute_slice_variance_inside_array(self, x):
        start = 10
        end = 30

        expected_mean = x[start:end].mean()
        expected_var = x[start:end].var()

        x_cumsum = np.cumsum(x)
        x2_cumsum = np.cumsum(x**2)
        actual_mean, actual_var = raw._compute_slice_stats(x_cumsum, x2_cumsum, start, end)

        assert np.isclose(expected_mean, actual_mean)
        assert np.isclose(expected_var, actual_var)

    def test_compute_slice_variance_start_array(self, x):
        start = 0
        end = 30

        expected_mean = x[start:end].mean()
        expected_var = x[start:end].var()

        x_cumsum = np.cumsum(x)
        x2_cumsum = np.cumsum(x**2)
        actual_mean, actual_var = raw._compute_slice_stats(x_cumsum, x2_cumsum, start, end)

        assert np.isclose(expected_mean, actual_mean)
        assert np.isclose(expected_var, actual_var)

    def test_compute_slice_variance_end_array(self, x):
        start = 0
        end = self.size

        expected_mean = x[start:end].mean()
        expected_var = x[start:end].var()

        x_cumsum = np.cumsum(x)
        x2_cumsum = np.cumsum(x**2)
        actual_mean, actual_var = raw._compute_slice_stats(x_cumsum, x2_cumsum, start, end)

        assert np.isclose(expected_mean, actual_mean)
        assert np.isclose(expected_var, actual_var)


class TestCombineSeedCandidates:
    def test_empty_data_ok(self):
        mz = np.array([])
        params = raw.MakeRoiParameters(tolerance=0.005)
        candidates = raw._combine_seed_candidates(mz, params)
        assert len(candidates) == 0

    def test_length_one_ok(self):
        mz = np.array([0.001])
        params = raw.MakeRoiParameters(tolerance=0.005)
        candidates = raw._combine_seed_candidates(mz, params)
        assert len(candidates) == 1

    def test_combine_ok_case1(self):
        mz = np.array([0.001, 0.002, 0.003, 0.5, 0.8, 0.801, 0.802])
        params = raw.MakeRoiParameters(tolerance=0.005)
        expected = np.array([0.002, 0.5, 0.801])
        actual = raw._combine_seed_candidates(mz, params)

        assert np.allclose(actual, expected)

    def test_combine_ok_case2(self):
        mz = np.array([0.001, 0.002, 0.003, 0.5, 0.8, 0.801, 0.802, 0.9])
        params = raw.MakeRoiParameters(tolerance=0.005)
        expected = np.array([0.002, 0.5, 0.801, 0.9])
        actual = raw._combine_seed_candidates(mz, params)

        assert np.allclose(actual, expected)

    def test_combine_large_array_ok(self):
        params = raw.MakeRoiParameters(tolerance=0.005)
        arr_list = list()
        size = 1000
        expected = np.zeros(size)
        for k in range(size):
            cluster = np.random.normal(size=10, loc=k, scale=params.tolerance / 2)
            expected[k] = np.mean(cluster)
            arr_list.append(cluster)
        mz = np.sort(np.concatenate(arr_list))
        actual = raw._combine_seed_candidates(mz, params)
        assert np.allclose(actual, expected)



"""Annotation tools configuration models."""

import pydantic

from ..chem import EnvelopeValidatorConfiguration


class AnnotatorParameters(EnvelopeValidatorConfiguration):
    """Store isotopologue annotator parameters."""

    max_charge: int = 3
    """Maximum charge of the features. Use negative values for negative polarity."""

    min_similarity: float = pydantic.Field(default=0.9, ge=0.5, le=1.0)
    """Minimum cosine similarity between a pair of features"""

    min_p: float = pydantic.Field(default=0.005, ge=0.0, le=0.1)
    """Minimum abundance of isotopes to include in candidate search."""



"""Utilities for working with chemical entities.

Refer to the :ref:`use guide <chemistry-user-guide>` for an introduction an examples on how to
use this package.

Constants
---------
- EM : the electron mass

"""

from .atoms import EM, Element, Isotope
from .config import EnvelopeScorerConfiguration, EnvelopeValidatorConfiguration, FormulaGeneratorConfiguration
from .context import DEFAULT_CONTEXT, ChemicalContext
from .envelope import EnvelopeScorer, EnvelopeValidator, score_envelope
from .formula import Formula
from .formula_generator import FormulaGenerator
from .table import PTABLE, PeriodicTable

__all__ = [
    "DEFAULT_CONTEXT",
    "EM",
    "PTABLE",
    "ChemicalContext",
    "Element",
    "EnvelopeScorerConfiguration",
    "EnvelopeValidatorConfiguration",
    "Formula",
    "FormulaGenerator",
    "FormulaGeneratorConfiguration",
    "Isotope",
    "PeriodicTable",
    "EnvelopeScorer",
    "EnvelopeValidator",
    "score_envelope",
]



"""Registry for concrete classes."""

from typing import Generic, TypeVar

from .exceptions import RegistryError, RepeatedIdError

T = TypeVar("T")


class Registry(Generic[T]):
    """Maintains a registry of related classes."""

    def __init__(self, name: str):
        self._name = name
        self._records: dict[str, type[T]] = dict()

    def get(self, id_: str) -> type[T]:
        """Retrieve a class from the registry."""
        if id_ not in self._records:
            raise RegistryError(f"Entry {id_} not found in {self._name} registry.")
        return self._records[id_]

    def register(self, *ids: str):
        """Add a class to the registry.

        Use as a decorator.

        """
        for id_ in ids:
            if id_ in self._records:
                raise RepeatedIdError(id_)

        def wrapper(entry: type[T]) -> type[T]:
            if not ids:
                self._records[entry.__name__] = entry
            else:
                for id_ in ids:
                    self._records[id_] = entry
            return entry

        return wrapper


operator_registry = Registry("operator")



import pathlib

import pytest

from tidyms2.core.exceptions import SampleNotFound
from tidyms2.lcms import sample_builder

from ..helpers import create_sample


class TestBuildFromYAML:
    def test_template_from_location_without_files_raise_error(self, tmp_path: pathlib.Path):
        data_dir = tmp_path / "data"
        data_dir.mkdir()
        template_file = tmp_path / "template.yaml"
        with pytest.raises(SampleNotFound):
            sample_builder.create_yaml_sample_template("*", template_file, data_dir)

    def test_create_template_and_read_returns_equal_samples(self, tmp_path: pathlib.Path):
        data_dir = tmp_path / "data"
        data_dir.mkdir()

        expected = list()
        for k in range(1, 6):
            s = create_sample(data_dir, k)
            expected.append(s)

        template_file = tmp_path / "template.yaml"

        sample_builder.create_yaml_sample_template("*", template_file, data_dir)

        actual = sample_builder.read_samples_from_yaml_template(template_file)
        assert actual == expected


class TestBuildFromCSV:
    def test_template_from_location_without_files_raise_error(self, tmp_path: pathlib.Path):
        data_dir = tmp_path / "data"
        data_dir.mkdir()
        template_file = tmp_path / "template.csv"
        with pytest.raises(SampleNotFound):
            sample_builder.create_csv_sample_template("*", template_file, data_dir)

    def test_create_template_and_read_returns_equal_samples(self, tmp_path: pathlib.Path):
        data_dir = tmp_path / "data"
        data_dir.mkdir()

        expected = list()
        for k in range(1, 6):  # start at order=1 so it matches order create by the builder function
            s = create_sample(data_dir, k)
            expected.append(s)

        template_file = tmp_path / "template.csv"

        sample_builder.create_csv_sample_template("*", template_file, data_dir)

        actual = sample_builder.read_samples_from_csv_template(template_file)
        assert actual == expected



import pathlib
from math import isclose

import numpy as np
import pydantic
import pytest

from tidyms2.core.models import MZTrace, Sample
from tidyms2.lcms.models import Peak


@pytest.fixture(scope="module")
def sample():
    return Sample(id="example", path=pathlib.Path("."))


class TestMZTrace:
    trace_size = 20

    @pytest.fixture
    def trace(self, sample) -> MZTrace:
        scans = np.arange(self.trace_size)
        time = np.linspace(0, 20, self.trace_size)
        spint = time.copy()
        mz = np.ones_like(time)
        return MZTrace(scan=scans, mz=mz, time=time, spint=spint, sample=sample)

    def test_noise_with_negative_values_raise_error(self, trace):
        noise = np.ones_like(trace.time)
        noise[self.trace_size // 2] = -0.005
        with pytest.raises(pydantic.ValidationError):
            trace.noise = noise

    def test_baseline_higher_than_intensity_raises_error(self, trace: MZTrace):
        baseline = trace.spint.copy()
        baseline[10] += 0.01
        with pytest.raises(pydantic.ValidationError):
            trace.baseline = baseline

    def test_serialize_deserialize_no_baseline(self, trace: MZTrace, sample):
        expected = trace
        actual = MZTrace.from_str(trace.to_str(), sample)
        assert actual.equals(expected)

    def test_serialize_deserialize_with_baseline_and_noise(self, trace: MZTrace, sample):
        expected = trace
        expected.noise = np.ones_like(trace.time)
        expected.baseline = trace.spint.copy()
        actual = MZTrace.from_str(trace.to_str(), sample)
        assert actual.equals(expected)


class TestPeak:
    trace_size = 201
    peak_start = 90
    peak_apex = 100
    peak_end = 111
    peak_mz = 200.0
    peak_height = 10.0
    noise_level = 2.0
    baseline_height = 1.0

    @pytest.fixture
    def trace(self, sample) -> MZTrace:
        """Use a square pulse as a LC-MS peak for unit testing."""
        scans = np.arange(self.trace_size)
        time = np.linspace(0, self.trace_size - 1, self.trace_size)

        mz = np.ones_like(time) * self.peak_mz
        signal = np.zeros_like(time)
        signal[self.peak_start : self.peak_end] += self.peak_height
        baseline = np.ones_like(time) * self.baseline_height
        trace_int = signal + baseline
        noise = np.ones_like(time) * self.noise_level
        return MZTrace(scan=scans, mz=mz, time=time, spint=trace_int, noise=noise, baseline=baseline, sample=sample)

    @pytest.fixture
    def peak(self, trace: MZTrace) -> Peak:
        return Peak(start=self.peak_start, apex=self.peak_apex, end=self.peak_end, roi=trace)

    @pytest.mark.parametrize(
        "start,apex,end", [(0, 0, 10), (10, 5, 12), (5, 10, 10), (5, 10, 8), (0, 10, trace_size + 1)]
    )
    def test_invalid_peak_definition_raises_error(self, trace, start, apex, end):
        with pytest.raises(pydantic.ValidationError):
            Peak(start=start, apex=apex, end=end, roi=trace)

    def test_serialize_deserialize(self, peak):
        expected = peak
        actual = Peak.from_str(expected.to_str(), expected.roi, expected.annotation)
        assert actual == expected

    def test_describe(self, peak):
        for k, v in peak.describe().items():
            assert isinstance(k, str)
            assert isinstance(v, float)

    def test_get_mz_ok(self, peak):
        assert isclose(peak.get("mz"), self.peak_mz)
        assert isinstance(peak.mz, float)

    def test_get_mz_when_intensity_is_equal_to_baseline_returns_nan(self, peak):
        peak.roi.baseline = peak.roi.spint
        assert np.isnan(peak.get("mz"))

    def test_get_rt_start_ok(self, peak):
        assert isclose(peak.get("rt_start"), peak.roi.time[peak.start])

    def test_get_rt_end_ok(self, peak):
        assert isclose(peak.get("rt_end"), peak.roi.time[peak.end - 1])

    def test_get_rt_ok(self, peak):
        assert isclose(peak.get("rt"), peak.roi.time[peak.apex])

    def test_get_rt_when_intensity_is_equal_to_baseline_returns_nan(self, peak):
        peak.roi.baseline = peak.roi.spint
        assert np.isnan(peak.get("rt"))

    def test_get_height(self, peak):
        assert isclose(peak.get("height"), self.peak_height)

    def test_get_height_no_baseline(self, peak):
        peak.roi.baseline = None
        assert isclose(peak.get("height"), self.peak_height + self.baseline_height)

    def test_get_area_ok(self, peak):
        expected_area = (peak.roi.time[self.peak_end - 1] - peak.roi.time[self.peak_start]) * self.peak_height
        assert isclose(expected_area, peak.get("area"))

    def test_get_area_ok_no_baseline(self, peak):
        expected_area = (peak.roi.time[self.peak_end - 1] - peak.roi.time[self.peak_start]) * (
            self.peak_height + self.baseline_height
        )
        peak.roi.baseline = None
        assert isclose(expected_area, peak.get("area"))

    def test_get_area_when_intensity_is_equal_to_baseline_returns_zero(self, peak):
        peak.roi.baseline = peak.roi.spint
        assert isclose(peak.get("area"), 0.0)

    def test_get_snr_ok(self, peak):
        assert isclose(peak.get("snr"), self.peak_height / self.noise_level)

    def test_get_snr_no_baseline_ok(self, peak):
        peak.roi.baseline = None
        assert peak.get("snr") > 0.0

    def test_get_snr_baseline_equal_to_intensity_returns_zero(self, peak):
        peak.roi.baseline = peak.roi.spint
        assert isclose(peak.get("snr"), 0.0)

    def test_get_snr_noise_level_equal_to_zero_returns_nan(self, peak):
        peak.roi.noise = np.zeros_like(peak.roi.time)
        assert np.isnan(peak.get("snr"))

    def test_get_snr_no_noise_level_returns_nan(self, peak):
        peak.roi.noise = None
        assert np.isnan(peak.get("snr"))

    def test_get_width_ok(self, peak):
        assert peak.get("width") > 0
        assert peak.get("width") <= peak.get("extension")

    def test_get_width_no_baseline_ok(self, peak):
        peak.roi.baseline = None
        assert peak.get("width") > 0
        assert peak.get("width") <= peak.get("extension")

    def test_get_width_baseline_equal_to_intensity_return_nan(self, peak):
        peak.roi.baseline = peak.roi.spint
        assert np.isnan(peak.get("width"))

    def test_compare_with_itself_returns_1(self, peak: Peak):
        actual = peak.compare(peak)
        expected = 1.0
        assert isclose(actual, expected)

    def test_compare_non_overlapping_peaks_return_0(self, peak: Peak):
        length = peak.end - peak.start + 1
        other = peak.model_copy()
        other.end += length
        other.apex += length
        other.start += length
        expected = 0.0
        assert isclose(peak.compare(other), expected)
        assert isclose(other.compare(peak), expected)

    @pytest.mark.parametrize("shift", [0, 1, 2, 3, 4, 5])
    def test_compare_is_symmetric(self, peak: Peak, shift):
        other = peak.model_copy()
        other.end += -shift
        other.start += shift
        similarity = peak.compare(other)
        assert isclose(similarity, other.compare(peak))
        assert 0.0 <= similarity
        assert (similarity < 1.0) or isclose(similarity, 1.0)

    @pytest.mark.parametrize("shift", [0, 1, 2, 3, 4, 5])
    def test_compare_is_in_valid_range(self, peak: Peak, shift):
        other = peak.model_copy()
        other.end += -shift
        other.start += shift
        similarity = peak.compare(other)
        assert 0.0 <= similarity
        assert (similarity < 1.0) or isclose(similarity, 1.0)



"""LC-MS operators."""

from .assay import LCFeatureMatcher
from .sample import LCPeakExtractor, LCTraceBaselineEstimator, LCTraceExtractor, LCTraceSmoother

__all__ = [
    "LCFeatureMatcher",
    "LCPeakExtractor",
    "LCTraceBaselineEstimator",
    "LCTraceExtractor",
    "LCTraceSmoother",
]



from pathlib import Path

import matplotlib.pyplot as plt

from tidyms2.core.models import Sample
from tidyms2.simulation.lcms import SimulatedLCMSAdductSpec, simulate_data_matrix

formula_list = ["[C54H104O6]+", "[C27H40O2]+", "[C24H26O12]+"]
rt_list = [10.0, 20.0, 30.0]
abundances = [1000.0, 2000.0, 3000.0]
adducts = list()
for rt, abundance, formula in zip(rt_list, abundances, formula_list):
    adduct = SimulatedLCMSAdductSpec(
        formula=formula,
        n_isotopologues=2,
        abundance={"mean": abundance, "std": 20.0}, # type: ignore
        response={"max_sensitivity_loss": 0.5, "sensitivity_decay": 0.25, "interbatch_variation": 0.5}, # type: ignore
    )
    adducts.append(adduct)


samples = list()
for k in range(40):
    id_ = f"sample-{k}"
    path = Path.cwd() / f"{id_}.mzML"
    s = Sample(id=id_, path=path)
    s.meta.order = k
    s.meta.batch = k // 10
    samples.append(s)

data = simulate_data_matrix(adducts, samples)

col = data.get_columns(0)[0]
order = [s.meta.order for s in data.list_samples()]

fig, ax = plt.subplots()

ax.plot(order, col.data, color="tab:grey")
ax.set_title(f"mz={col.feature.mz:.4f}, Rt={col.feature.descriptors["rt"]:.1f} s")
ax.fill_between([0, 9], 100, 500, alpha=0.25, label="batch=1")
ax.fill_between([10, 19], 100, 500, alpha=0.25, label="batch=2")
ax.fill_between([20, 29], 100, 500, alpha=0.25, label="batch=3")
ax.fill_between([30, 39], 100, 500, alpha=0.25, label="batch=4")
ax.legend()


ax.set_xlabel("Run order")
ax.set_ylabel("Intensity [au]")
ax.set_ylim(0, 600.0)



"""Storage tools for the annotation algorithm."""

from typing import Generic, Sequence

from ..core.models import AnnotableFeatureType


class AnnotationData(Generic[AnnotableFeatureType]):
    """Container class for data used during isotopologue annotation.

    :param features: list of features sorted by m/z
    :param similarity_cache: stores similarity between features.
    :param non_annotated : non-annotated features.

    """

    def __init__(self, features: Sequence[AnnotableFeatureType]):
        self.features = sorted(features, key=lambda x: x.mz)
        self.non_annotated = set(features)
        self._monoisotopologues = sorted(features, key=lambda x: x.height)
        self.similarity_cache = SimilarityCache()
        self._label_counter = 0

    def get_monoisotopologue(self) -> AnnotableFeatureType | None:
        """Get the current non-annotated feature with the greatest area."""
        if self._monoisotopologues:
            mono = self._monoisotopologues[-1]
            while self._monoisotopologues and (mono not in self.non_annotated):
                self._monoisotopologues.pop()
                if self._monoisotopologues:
                    mono = self._monoisotopologues[-1]
                else:
                    mono = None
        else:
            mono = None
        return mono

    def annotate(self, features: Sequence[AnnotableFeatureType], charge: int):
        """Labels a list of features as an isotopic envelope."""
        if len(features) > 1:
            for k, ft in enumerate(features):
                assert ft.annotation is not None
                ft.annotation.charge = charge
                ft.annotation.isotopologue_label = self._label_counter
                ft.annotation.isotopologue_index = k
                self._flag_annotated(ft)
            self._label_counter += 1
        else:
            self._flag_annotated(features[0])

    def _flag_annotated(self, feature: AnnotableFeatureType):
        """Flag features as annotated."""
        self.non_annotated.discard(feature)
        if self._monoisotopologues and (feature == self._monoisotopologues[-1]):
            self._monoisotopologues.pop()


class SimilarityCache(Generic[AnnotableFeatureType]):
    """Stores and retrieves the similarity between features in a sample."""

    def __init__(self):
        self._cache: dict[AnnotableFeatureType, dict[AnnotableFeatureType, float]] = dict()

    def get_similarity(self, ft1: AnnotableFeatureType, ft2: AnnotableFeatureType):
        """Get the similarity between a pair of features."""
        ft1_sim = self._cache.setdefault(ft1, dict())
        ft2_sim = self._cache.setdefault(ft2, dict())
        if ft2 in ft1_sim:
            similarity = ft1_sim[ft2]
        else:
            similarity = ft1.compare(ft2)
            ft1_sim[ft2] = similarity
            ft2_sim[ft1] = similarity
        return similarity



"""Periodic table implementation."""

import json
import pathlib
from string import digits

from .atoms import Element, Isotope
from .exceptions import IsotopeError


class PeriodicTable:
    """Store information from elements and their isotopes.

    :param custom_abundances: override natural abundances of isotopes.
        Custom abundances is a dictionary that maps atomic numbers to a dictionary
        of isotopes mass numbers and their abundance. If an abundance is not
        provided for a particular isotope it is assumed to be zero. Abundances from
        a set of isotopes must be normalized to one.

    """

    def __init__(self, custom_abundances: dict[str, dict[int, float]] | None = None):
        self._symbol_to_element = _make_periodic_table(custom_abundances)
        self._z_to_element = {v.z: v for v in self._symbol_to_element.values()}
        self._za_to_isotope = dict()
        self._str_to_isotope = dict()
        for el_str in self._symbol_to_element:
            el = self._symbol_to_element[el_str]
            for isotope in el.isotopes:
                self._za_to_isotope[(isotope.z, isotope.a)] = isotope
                self._str_to_isotope[str(isotope.a) + el_str] = isotope

    def get_element(self, element: str | int) -> Element:
        """Fetch an element object using its symbol or atomic number.

        >>> import tidyms as ms
        >>> ptable = ms.chem.PeriodicTable()
        >>> h = ptable.get_element("H")
        >>> c = ptable.get_element(6)

        """
        if isinstance(element, int):
            res = self._z_to_element[element]
        else:
            res = self._symbol_to_element[element]
        return res

    def __iter__(self):
        for el in self._symbol_to_element.values():
            yield el

    def get_isotope(self, symbol: str, copy: bool = False) -> Isotope:
        """Fetch an isotope object from a string representation.

        :param symbol: a string representation of the isotope. If only the symbol
            is provided in the string, the monoisotope is returned.
        :param copy: If set to ``True`` a new isotope instance is created.

        >>> import tidyms as ms
        >>> ptable = ms.chem.PeriodicTable()
        >>> d = ptable.get_isotope("2H")
        >>> cl35 = ptable.get_isotope("Cl")

        """
        try:
            if symbol[0] in digits:
                isotope = self._str_to_isotope[symbol]
            else:
                isotope = self.get_element(symbol).monoisotope

            if copy:
                isotope = isotope.model_copy()

            return isotope
        except KeyError as e:
            msg = f"{symbol} is not a valid element or isotope symbol."
            raise IsotopeError(msg) from e

    def is_monoisotope(self, isotope: Isotope) -> bool:
        """Check if an isotope is the most abundant."""
        element = self.get_element(isotope.symbol)
        return element.monoisotope.a == isotope.a


def _make_periodic_table(custom_abundances: dict[str, dict[int, float]] | None = None) -> dict[str, Element]:
    data_dir = pathlib.Path(__file__).parent
    elements_file = data_dir / "elements.json"
    elements_dict = json.loads(elements_file.read_text())

    isotopes_file = data_dir / "isotopes.json"
    isotopes_dict = json.loads(isotopes_file.read_text())

    if custom_abundances is not None:
        _patch_isotope_abundances(isotopes_dict, custom_abundances)

    periodic_table = dict()
    for symbol, isotopes_list in isotopes_dict.items():
        name = elements_dict[symbol]
        isotopes = tuple(Isotope(symbol=symbol, **x) for x in isotopes_list)
        z = isotopes_list[0]["z"]
        element = Element(symbol=symbol, name=name, z=z, isotopes=isotopes)
        periodic_table[symbol] = element

    return periodic_table


def _patch_isotope_abundances(isotopes: dict[str, list[dict]], custom_abundances: dict[str, dict[int, float]]) -> None:
    for symbol, element_isotopes in isotopes.items():
        isotope_patches = custom_abundances.get(symbol)
        if isotope_patches is None:
            continue
        for i in element_isotopes:
            i["p"] = isotope_patches.get(i["a"], 0.0)


PTABLE = PeriodicTable()



"""A manager class for processing multiple samples."""

from collections import OrderedDict
from logging import getLogger
from typing import Generic, Iterable, Literal, overload
from uuid import UUID

from ..core import exceptions
from ..core.matrix import DataMatrix
from ..core.models import Annotation, FeatureType, RoiType, Sample
from ..core.operators.assay import AssayOperator, MissingImputer
from ..core.operators.pipeline import Pipeline
from ..core.storage import AssayStorage, SampleStorage
from .executors import SampleProcessor
from .utils import create_feature_groups, create_matrix_data

logger = getLogger("assay")


class Assay(Generic[RoiType, FeatureType]):
    """The assay class.

    :param id: an identifier for the assay
    :param assay_storage: the storage class for assay data.

    """

    class _PipelineContainer:
        def __init__(self, id: str):
            self.sample = Pipeline(f"{id}-sample-pipeline")
            self.assay = Pipeline(f"{id}-assay-pipeline")

    def __init__(
        self,
        id: str,
        assay_storage: AssayStorage[RoiType, FeatureType],
        sample_processor: SampleProcessor[RoiType, FeatureType],
        sample_storage_type: type[SampleStorage[RoiType, FeatureType]],
        sample_storage_config: dict | None = None,
    ):
        self.id = id
        self._sample_queue: OrderedDict[str, Sample] = OrderedDict()
        self._storage = assay_storage
        self._sample_processor = sample_processor
        self._sample_storage_type = sample_storage_type
        self._sample_storage_config = sample_storage_config or dict()
        self.pipes = self._PipelineContainer(id)

    def add_samples(self, *samples: Sample) -> None:
        """Add samples to the assay sample processing queue.

        :param sample: the samples to add
        """
        # first we validate new samples
        Sample.validate_samples(*samples)

        # then we also include processed samples
        all_samples = self._storage.list_samples() + list(samples)
        try:
            Sample.validate_samples(*all_samples)
        except exceptions.RepeatedIdError as e:
            msg = "Cannot add samples. Processed samples with the same id already exist."
            raise exceptions.RepeatedIdError(msg) from e
        except exceptions.RepeatedSampleOrder as e:
            msg = "Cannot add samples. Processed samples with the run order already exist."
            raise exceptions.RepeatedIdError(msg) from e

        for sample in samples:
            logger.info(f"Added sample `{sample.id}` with path `{sample.path}` to {self.id}.")
            self._sample_queue[sample.id] = sample

    def create_data_matrix(self, matrix_value: str | None = None) -> DataMatrix:
        """Create a data matrix using data from the assay pipeline.

        :param matrix_value: the descriptor name used to build the matrix data. If not
            provided, the descriptor `area` will be used as the matrix value.

        """
        matrix_value = matrix_value or "area"
        descriptors = self.fetch_feature_table()
        if matrix_value not in descriptors:
            raise ValueError(f"{matrix_value} is not a valid feature descriptor.")
        values = descriptors[matrix_value]
        groups = self._storage.fetch_feature_groups()
        samples = self.fetch_samples()
        annotations = self.fetch_feature_annotations()
        fill_values = self._storage.fetch_fill_values()

        data = create_matrix_data(values, annotations, samples, fill_values)
        return DataMatrix(samples, groups, data)

    def fetch_samples(self, queued: bool = False) -> list[Sample]:
        """Retrieve a list of queued or processed samples in the assay.

        :param queued: if set to ``True``, return queued samples. Otherwise, fetch processed samples.

        """
        if queued:
            return [x for x in self._sample_queue.values()]

        return self._storage.list_samples()

    def fetch_feature_annotations(self) -> list[Annotation]:
        """Retrieve a list with feature annotations."""
        return self._storage.fetch_annotations()

    def fetch_feature_table(self) -> dict[str, list[float]]:
        """Fetch the feature descriptors table."""
        return self._storage.fetch_descriptors()

    @overload
    def fetch_features(self, by: Literal["id"], keys: Iterable[UUID]) -> list[FeatureType]: ...

    @overload
    def fetch_features(self, by: Literal["sample"], keys: str) -> list[FeatureType]: ...

    @overload
    def fetch_features(self, by: Literal["group"], keys: int) -> list[FeatureType]: ...

    def fetch_features(self, by: str, keys) -> list[FeatureType]:
        """Fetch extracted features."""
        if by == "id":
            return self._storage.fetch_features_by_id(*keys)
        elif by == "sample":
            return self._storage.fetch_features_by_sample(keys)
        elif by == "group":
            return self._storage.fetch_features_by_group(keys)
        else:
            raise ValueError(f"Valid values for `by` are `sample`, 'group' or `id`. Got {by}.")

    @overload
    def fetch_rois(self, by: Literal["id"], keys: Iterable[UUID]) -> list[RoiType]: ...

    @overload
    def fetch_rois(self, by: Literal["sample"], keys: str) -> list[RoiType]: ...

    def fetch_rois(self, by: str, keys):
        """Fetch extracted ROIs.

        :param by: The criteria to fetch ROIs, can be either ``"id"`` or ``"sample"``.
        :param keys: the keys used to fetch ROIs. If `by` is set to ``"id"``, then keys must be a list of
            ROI ids. If `by` is set to ``"sample"`` it must be a single sample id.

        """
        if by == "sample":
            return self._storage.fetch_rois_by_sample(keys)
        elif by == "id":
            return self._storage.fetch_rois_by_id(*keys)
        else:
            raise ValueError(f"Valid values for parameter `by` are `sample` or `id`. Got {by}.")

    def process_samples(self, sample_ids: list[str] | None = None) -> None:
        """Apply sample pipeline to queued samples.

        :param sample_ids: process only a subset of samples in the queue. This parameter is useful
            to explore sample processing parameters. If not provided, process all samples in the
            queue.

        """
        if not self._sample_queue:
            logger.warning("No samples to process: sample queue is empty.")
            return
        if sample_ids is None:
            sample_ids = list(self._sample_queue)

        samples = list()
        for id_ in sample_ids:
            try:
                sample = self._sample_queue[id_]
                sample_storage = self._sample_storage_type(
                    sample,
                    self._storage.get_roi_type(),
                    self._storage.get_feature_type(),
                    **self._sample_storage_config,
                )
                samples.append(sample_storage)
            except KeyError:
                raise exceptions.SampleNotFound(id_)

        logger.info("Starting sample processing pipeline with operators:")
        for k, op in enumerate(self.pipes.sample.operators, start=1):
            logger.info(f"{k}. {op.id}: {op}")
        self._sample_processor.execute(self._storage, self.pipes.sample, *samples)
        self._remove_from_sample_queue()
        logger.info("Successfully processed all samples.")

    def process_assay(self) -> None:
        """Apply assay pipeline to assay data."""
        if self._sample_queue:
            msg = "There are unprocessed samples in the sample queue. Run `process_samples` to fix this error."
            raise exceptions.UnprocessedSampleError(msg)

        feature_groups_created = False
        for op in self.pipes.assay.operators:
            assert isinstance(op, AssayOperator)  # helping pyright
            if isinstance(op, MissingImputer):
                self._create_feature_groups()
                feature_groups_created = True
            logger.info(f"Applying {op.id} to assay...")
            op.apply(self._storage)

        if not feature_groups_created:
            self._create_feature_groups()

    def _remove_from_sample_queue(self, ids: list[str] | None = None) -> None:
        if ids is None:
            self._sample_queue.clear()
        else:
            for id_ in ids:
                self._sample_queue.pop(id_)

    def _create_feature_groups(self) -> None:
        descriptors = self.fetch_feature_table()
        annotations = self.fetch_feature_annotations()
        feature_groups = create_feature_groups(descriptors, annotations)
        self._storage.add_feature_groups(*feature_groups)



"""Implementation of the Minimum Mass Finder."""

import bisect

import numpy as np

from ..chem import EM, ChemicalContext
from ..chem.atoms import Element
from ..chem.envelope import make_formula_coefficients_envelopes
from ..chem.formula_generator import FormulaCoefficientBounds
from ..core.models import AnnotableFeatureType
from .annotation_data import AnnotationData
from .config import AnnotatorParameters


class MMIFinder:
    """Find the Minimum Mass Isotopologue (MMI) in a list of features.

    Parameters
    ----------
    bounds : dict
        Mapping from an element symbol str to the minimum and maximum
        allowed values in formulas.
    max_mass : float
        Maximum mass to build rules.
    length : int
        length of the theoretical envelopes used to compute the search
        rules.
    bin_size : int
        Mass interval used to build the rules.
    mz_tol : float
        m/z tolerance to search candidates.
    p_tol : float
        abundance tolerance used to search candidates.
    min_similarity : float, default=0.9
        Minimum similarity to create candidates.
    context: chemical context

    """

    def __init__(self, config: AnnotatorParameters, bin_size: int, context: ChemicalContext):
        self.context = context
        self.rules = _create_rules_dict(config, bin_size, context)
        self.bin_size = bin_size
        self.max_charge = abs(config.max_charge)
        self.polarity = 1 if config.max_charge >= 0 else -1
        self.config = config

    def find(self, data: AnnotationData[AnnotableFeatureType]) -> list[tuple[AnnotableFeatureType, int]]:
        """Search MMI candidates using features m/z and area.

        :param features: feature data to search.
        :return: list of MMI candidates and their possible charge.

        """
        mono = data.get_monoisotopologue()
        candidates = list()

        if mono is None:
            return candidates

        for charge in range(1, self.max_charge + 1):
            M_mono = mono.mz * charge - self.polarity * charge * EM
            if M_mono < self.config.max_M:
                candidates.append((mono, charge))
            M_bin = int(M_mono // self.bin_size)
            mmi_rules = self.rules.get(M_bin)
            if mmi_rules is not None:
                for i_rules in mmi_rules:
                    i_candidates = _find_candidate(data, mono, charge, i_rules, self.config)
                    candidates.extend(i_candidates)
        return candidates


def _find_candidate(
    data: AnnotationData[AnnotableFeatureType],
    mono: AnnotableFeatureType,
    charge: int,
    i_rules: dict,
    config: AnnotatorParameters,
) -> list[tuple[int, int]]:
    # search valid m/z values
    min_dM, max_dM = i_rules["dM"]
    min_mz = mono.mz - max_dM / charge - config.max_M_tol
    max_mz = mono.mz - min_dM / charge + config.max_M_tol
    min_qp = i_rules["qp"][0] - config.p_tol
    max_qp = i_rules["qp"][1] + config.p_tol

    if (mono.mz * charge) < config.max_M:
        start = bisect.bisect(data.features, min_mz, key=lambda x: x.mz)
        end = bisect.bisect(data.features, max_mz, key=lambda x: x.mz)
    else:
        start, end = 0, 0  # dummy values

    # if valid m/z where found, check if the abundance quotient qp is valid
    candidates = list()
    if start < end:
        for k in range(start, end):
            candidate = data.features[k]
            is_valid = _check_candidate(data, mono, candidate, config.min_similarity, min_qp, max_qp)
            if is_valid:
                candidates.append((candidate, charge))
    return candidates


def _check_candidate(
    data: AnnotationData[AnnotableFeatureType],
    mono: AnnotableFeatureType,
    candidate: AnnotableFeatureType,
    min_similarity: float,
    min_qp: float,
    max_qp: float,
) -> bool:
    if candidate not in data.non_annotated:
        return False

    similarity = data.similarity_cache.get_similarity(mono, candidate)

    if similarity < min_similarity:
        return False

    p = mono.compute_isotopic_envelope(candidate, mono).p
    qp = p[1] / p[0]
    is_valid_qp = (qp >= min_qp) & (qp <= max_qp)

    return is_valid_qp


def _create_rules_dict(
    config: AnnotatorParameters, bin_size: int, context: ChemicalContext
) -> dict[int, list[dict[str, tuple[float, float]]]]:
    Ma, pa = _create_envelope_arrays(config, context)
    # find the monoisotopic index, its Mass difference with the MMI (dM) and
    # its abundance quotient with the MMI (qp)
    bins = (Ma[:, 0] // bin_size).astype(int)

    # find unique values for bins and monoisotopic index that will be used
    # as key for the rule dictionary
    unique_bins = np.unique(bins)
    # unique_mono_index = np.unique(mono_index)
    # unique_mono_index = unique_mono_index[unique_mono_index > 0]

    rules = dict()
    for b in unique_bins:
        b_rules = list()
        bin_mask = bins == b
        for mi in range(1, config.max_length):
            qp = pa[bin_mask, mi] / pa[bin_mask, 0]
            dM = Ma[bin_mask, mi] - Ma[bin_mask, 0]
            qp_mask = qp >= (1.0 - config.p_tol)
            if qp_mask.any():
                mi_rules = dict()
                dM_b_mi = dM[qp_mask]
                qp_b_mi = qp[qp_mask]
                mi_rules["dM"] = dM_b_mi.min(), dM_b_mi.max()
                mi_rules["qp"] = qp_b_mi.min(), qp_b_mi.max()
                b_rules.append(mi_rules)
        if b_rules:
            rules[b] = b_rules
    return rules


def _create_envelope_arrays(config: AnnotatorParameters, context: ChemicalContext) -> tuple[np.ndarray, np.ndarray]:
    selected_elements = {x.symbol for x in _select_elements([context.table.get_element(x) for x in config.bounds])}
    bounds = {k: v for k, v in config.bounds.items() if k in selected_elements}
    coeff_bounds = FormulaCoefficientBounds.from_isotope_str(bounds)
    formula_coefficients = coeff_bounds.make_coefficients(config.max_M)
    envelope = make_formula_coefficients_envelopes(
        bounds, formula_coefficients, context.envelope_cache, context.table, config.max_length
    )
    M = envelope.M
    p = envelope.p
    return M, p


def _select_two_isotope_element(elements: list[Element], dm: int) -> list[Element]:
    """Select elements from the list with two isotopes.

    All elements where the abundance of the MMI is lower than the other isotope are kept.
    In cases where the MMI is greater than the other isotope, only the isotope with
    the minimum abundance from this subset is kept.

    """
    selected: list[Element] = list()
    max_p1_where_p0_gt_p1 = 0.0  # maximum value of two element isotopes p0 where p0 > p1
    best_p0_greater_than_p1 = None
    for element in elements:
        if len(element.isotopes) != 2:
            continue

        other = element.isotopes[1]

        element_dm = other.a - element.mmi.a
        if element_dm != dm:
            continue

        if other.p >= element.mmi.p:
            selected.append(element)
        elif other.p > max_p1_where_p0_gt_p1:
            max_p1_where_p0_gt_p1 = other.p
            best_p0_greater_than_p1 = element

    if best_p0_greater_than_p1 is not None:
        selected.append(best_p0_greater_than_p1)
    return selected


def _select_multiple_isotope_elements(e_list: list[Element]) -> list[Element]:
    selected = list()
    for e in e_list:
        n_isotopes = len(e.isotopes)
        if n_isotopes > 2:
            selected.append(e)
    return selected


def _select_elements(e_list: list[Element]) -> list[Element]:
    two_isotope_dm1 = _select_two_isotope_element(e_list, 1)
    two_isotope_dm2 = _select_two_isotope_element(e_list, 2)
    selected = _select_multiple_isotope_elements(e_list)
    selected.extend(two_isotope_dm1)
    selected.extend(two_isotope_dm2)
    return selected



"""Utilities to simulate MS data."""

from .base import BaseChemicalSpeciesSpec, DataAcquisitionSpec, MZGridSpec
from .lcms import SimulatedLCMSAdductSpec, SimulatedLCMSSampleFactory

__all__ = [
    "BaseChemicalSpeciesSpec",
    "MZGridSpec",
    "SimulatedLCMSAdductSpec",
    "SimulatedLCMSSampleFactory",
    "DataAcquisitionSpec",
]



"""LC-MS missing imputer."""

from __future__ import annotations

import bisect
import concurrent.futures
from logging import getLogger
from typing import Generator

from scipy.integrate import trapezoid

from ..algorithms.raw import MakeChromatogramParameters, make_chromatograms
from ..core.models import FeatureGroup, FillValue, Sample
from ..core.storage import AssayStorage
from ..io import MSData

logger = getLogger(__name__)


def compute_fill_values_lc(data: AssayStorage, mz_tolerance: float, max_workers: int | None = None) -> list[FillValue]:
    r"""Compute fill missing values using the chromatogram area.

    :param data:
        Assay data with features matched
    :param mz_tolerance:
        m/z tolerance used to build chromatograms
    :param max_workers:
        maximum number of workers for parallel execution
    :return:
        a list of fill values.

    Notes
    -----
    Missing Chromatographic peaks are searched in the raw data using the
    peak descriptors obtained after features extraction and correspondence.
    Initially, chromatograms for missing features in each sample are built by
    using the mean m/z from the features descriptors. A time window for
    searching the peaks is defined based on the mean Rt and the standard
    deviation values features:

    .. math::

        t_{mean} \pm n_{dev} t_{std}

    If a peak is found in the region, its area is used as fill value. If more
    than one peak is found in the region, the closes to the mean Rt is chosen.
    If no peak was found in the region, an estimation of the peak area is
    computed by integrating the region comprised by the mean start and end
    times of the peak in the detected features.

    """
    all_samples = data.list_samples()
    all_groups = data.fetch_feature_groups()

    def iterator(
        samples: list[Sample], groups: list[FeatureGroup]
    ) -> Generator[tuple[Sample, list[FeatureGroup]], None, None]:
        sample_id_to_sample = {x.id: x for x in samples}
        group_id_to_group = {x.group: x for x in groups}
        sample_to_groups = {x: set(group_id_to_group) for x in sample_id_to_sample}

        for ann in data.fetch_annotations():
            if ann.group == -1:
                continue

            sample_groups = sample_to_groups.get(ann.sample_id)
            if sample_groups is None:
                continue

            sample_groups.remove(ann.group)

        for sample_id, missing_groups in sample_to_groups.items():
            yield sample_id_to_sample[sample_id], [group_id_to_group[x] for x in missing_groups]

    result = list()

    params = MakeChromatogramParameters(window=mz_tolerance, fill_missing=True, mz=list())
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
        n_samples = len(all_samples)
        futures = [
            executor.submit(_worker, sample, group, params) for sample, group in iterator(all_samples, all_groups)
        ]
        for k, future in enumerate(concurrent.futures.as_completed(futures)):
            sample_fill_values = future.result()
            logger.info(f"Processing {k}/{n_samples}).")
            result.append(sample_fill_values)
    return result


def _worker(
    sample: Sample,
    features: list[FeatureGroup],
    params: MakeChromatogramParameters,
) -> list[FillValue]:
    """Find fill values for features not detected during feature extraction in a sample."""
    ms_data = MSData(sample)
    params = params.model_copy()
    params.mz = [x.mz for x in features]
    chromatograms = make_chromatograms(ms_data, params)

    fill_values = list()
    for chrom, group in zip(chromatograms, features):
        start = bisect.bisect_left(chrom.time, group.descriptors["rt_start"])
        end = bisect.bisect_left(chrom.time, group.descriptors["rt_end"])
        value = trapezoid(chrom.int[start:end], chrom.time[start:end])
        fill = FillValue(sample_id=sample.id, feature_group=group.group, value=value)
        fill_values.append(fill)
    return fill_values



"""Low level utilities to work with raw data."""

from __future__ import annotations

from collections import deque
from copy import deepcopy
from typing import Sequence

import numpy as np
import pydantic
from scipy.interpolate import interp1d

from ..core.enums import MSDataMode
from ..core.models import Chromatogram, MSSpectrum, MZTrace, Sample
from ..core.utils.numpy import FloatArray1D, IntArray1D, find_closest
from ..io.msdata import MSData


class BaseRawMsFunctionParameter(pydantic.BaseModel):
    """Base class that other parameter models inherit from."""

    ms_level: pydantic.PositiveInt | None = None
    """the data MS level"""

    start_time: pydantic.NonNegativeFloat | None = None
    """Use spectra with acquisition times greater than this value"""

    end_time: pydantic.NonNegativeFloat | None = None
    """Use spectra with acquisition times greater than this value. If ``None``, it doesn't
    filter scans by time.
    """


class MakeTICParameters(BaseRawMsFunctionParameter):
    """Store make_tic parameters."""

    kind: str = "tic"
    """`tic` computes the total ion chromatogram. `bpi` computes the base peak chromatogram"""

    @pydantic.field_validator("kind", mode="before")
    @classmethod
    def _check_kind(cls, value: str) -> str:
        msg = f"`kind` must be set either to `tic` or `bpi`. Got {value}."
        assert value in {"tic", "bpi"}, msg
        return value


class MakeChromatogramParameters(BaseRawMsFunctionParameter):
    """Store make_chromatogram parameters."""

    mz: Sequence[float]
    """a sorted sequence of m/z values"""

    window: pydantic.PositiveFloat = 0.05
    """m/z tolerance to build EICs."""

    accumulator: str = "sum"
    """How multiple values in a m/z windows are accumulated. ``"sum"`` computes
    the total intensity inside the window. ``"mean"`` divides the total intensity
    using the number of points inside the window.
    """

    fill_missing: bool = True
    """If ``True``, sets the intensity to zero if no signal was found in a m/z window.
    If ``False``, missing values are set to ``nan``.
    """


class MakeRoiParameters(BaseRawMsFunctionParameter):
    """Store make_roi parameters."""

    tolerance: float = 0.01
    """connect m/z values across scans if they are closer than this value"""

    max_missing: pydantic.NonNegativeInt = 1
    """maximum number of consecutive missing values in a ROI."""

    min_length: pydantic.PositiveInt = 5
    """The minimum length of a ROI, defined as the number of non-NaN values in the ROI."""

    min_intensity: pydantic.NonNegativeFloat = 0.0
    """Discard ROIs if all of its elements have an intensity lower than this value."""

    multiple_match: str = "reduce"
    """How points are matched when there are multiple matches. If ``"closest"``
    is used, the closest peak is assigned as a match and the others are used to
    create new ROIs. If ``"reduce"`` is used, an m/z and intensity pair is
    computed for all matching points using the mean for m/z and the sum for the
    intensity.
    """

    targeted_mz: Sequence[float] | None = None
    """If provided, only ROI with these m/z values will be created."""

    pad: pydantic.NonNegativeInt = 0
    """The number of dummy values to pad the ROI with"""

    @pydantic.field_validator("multiple_match", mode="before")
    @classmethod
    def _check_multiple_match(cls, value: str) -> str:
        msg = f"`multiple_match` must be set either to `closest` or `reduce`. Got {value}."
        assert value in {"closest", "reduce"}, msg
        return value


class AccumulateSpectraParameters(pydantic.BaseModel):
    """Store accumulate_spectra parameters."""

    ms_level: pydantic.PositiveInt = 1
    """the data MS level"""

    start_time: pydantic.NonNegativeFloat
    """Accumulate spectra starting at this acquisition time"""

    end_time: pydantic.NonNegativeFloat
    """Accumulate spectra starting until this acquisition time"""

    subtract_left_time: float | None = None
    """Scans with acquisition times lower than this value are subtracted from the
    accumulated spectrum. If ``None``, no subtraction is done.

    """
    subtract_right_time: float | None = None
    """Scans with acquisition times greater than this value are subtracted from the
    accumulated spectrum. If ``None``, no subtraction is done.
    """


def make_tic(data: MSData, params: MakeTICParameters) -> Chromatogram:
    """Create a total ion chromatogram.

    :param data: The data file used to create the TIC
    :param params: make TIC parameters.

    .. seealso:: MakeTICParameters

    """
    rt = list()
    tic = list()

    reduce = np.sum if params.kind == "tic" else np.max

    with data.using_tmp_config(ms_level=params.ms_level, start_time=params.start_time, end_time=params.end_time):
        for sp in data:
            rt.append(sp.time)
            tic.append(reduce(sp.int) if sp.int.size else 0.0)
    return Chromatogram(time=np.array(rt), int=np.array(tic))


def make_chromatograms(data: MSData, params: MakeChromatogramParameters) -> list[Chromatogram]:
    """Compute multiple :term:`EIC` from raw data using a list of m/z values.

    :param ms_data: raw data
    :param params: function parameters

    .. seealso:: MakeChromatogramParameters

    """
    n_sp = data.get_n_spectra()

    mz_arr = np.array(sorted(params.mz))

    # create an array of m/z windows: [mz0 - w, mz0 + w, mz1 - w, mz1 + w, ..., mzn + w]
    # where w is the window parameter
    mz_intervals = np.vstack((mz_arr - params.window, mz_arr + params.window)).T.reshape(mz_arr.size * 2)

    eic = np.zeros((mz_arr.size, n_sp))
    if not params.fill_missing:
        eic[:] = np.nan

    rt = np.zeros(n_sp)
    valid_index = list()
    with data.using_tmp_config(ms_level=params.ms_level, start_time=params.start_time, end_time=params.end_time):
        for sp in data:
            valid_index.append(sp.index)
            rt[sp.index] = sp.time

            # prevents error when working with empty spectra
            if sp.mz.size == 0:
                continue

            # consecutive values in window index defines the location of an m/z window
            window_index = np.searchsorted(sp.mz, mz_intervals)
            window_width = window_index[1::2] - window_index[::2]
            window_has_mz = (window_width) > 0
            # elements added at the end of mz_sp raise IndexError
            window_index[window_index >= sp.mz.size] = sp.mz.size - 1
            # Fast sum of values in a window
            fill = 0.0 if params.fill_missing else np.nan
            tmp_eic = np.where(window_has_mz, np.add.reduceat(sp.int, window_index)[::2], fill)

            if params.accumulator == "mean":
                window_width[~window_has_mz] = 1
                tmp_eic = tmp_eic / window_width
            eic[:, sp.index] = tmp_eic

    valid_index = np.array(valid_index)
    rt = rt[valid_index]
    eic = eic[:, valid_index]

    chromatograms = list()
    for row in eic:
        chromatogram = Chromatogram(time=rt.copy(), int=row)
        chromatograms.append(chromatogram)
    return chromatograms


def make_roi(data: MSData, params: MakeRoiParameters) -> list[MZTrace]:
    """Build regions of interest (ROI) from raw data.

    ROI are created by connecting values across scans based on the closeness in
    m/z. See the :ref:`algorithms-roi-extraction` for a description of the
    algorithm used.

    :param data : raw data
    :param params: algorithm parameters

    .. seealso:: lcms.MZTrace
    .. seealso:: MakeRoiParameters

    """
    if data.sample.ms_data_mode is MSDataMode.PROFILE and data.centroider is None:
        msg = "`ms_data` needs to be in centroid mode or needs to have a centroider."
        raise ValueError(msg)

    targeted = params.targeted_mz is not None

    if params.targeted_mz is not None:
        roi_seeds = np.sort(list(params.targeted_mz))
    else:
        if params.min_intensity > 0.0:
            roi_seeds = _compute_roi_seeds(data, params)
        else:
            roi_seeds = None

    rt = np.zeros(data.get_n_spectra())
    roi_maker = _RoiMaker(data.get_sample(), params, roi_seeds=roi_seeds, targeted=targeted)

    scans = list()
    with data.using_tmp_config(ms_level=params.ms_level, start_time=params.start_time, end_time=params.end_time):
        for spectrum in data:
            rt[spectrum.index] = spectrum.time
            scans.append(spectrum.index)
            roi_maker.feed_spectrum(spectrum)
            roi_maker.clear_completed_roi()

    # add roi not completed during the last scan
    roi_maker.flag_as_completed()
    roi_maker.clear_completed_roi()
    scans = np.array(scans)
    roi_list = roi_maker.tmp_roi_to_roi(scans, rt, params.pad)

    return roi_list


def accumulate_spectra(ms_data: MSData, params: AccumulateSpectraParameters) -> MSSpectrum:
    """Accumulate consecutive spectra into a single spectrum.

    :param ms_data: raw data file
    :param params: algorithm parameters

    .. seealso:: AccumulateSpectraParameters

    """
    if params.subtract_left_time is None:
        params.subtract_left_time = params.start_time

    if params.subtract_right_time is None:
        params.subtract_right_time = params.end_time

    if ms_data.sample.ms_data_mode is MSDataMode.CENTROID:
        return _accumulate_spectra_centroid(ms_data, params)
    return _accumulate_spectra_profile(ms_data, params)


def _accumulate_spectra_centroid(ms_data: MSData, params: AccumulateSpectraParameters) -> MSSpectrum:
    # don't remove any m/z value when detecting rois
    max_missing = ms_data.get_n_spectra()

    roi_params = MakeRoiParameters(
        max_missing=max_missing,
        min_length=1,
        start_time=params.subtract_left_time,
        end_time=params.subtract_right_time,
        ms_level=params.ms_level,
    )

    roi = make_roi(ms_data, roi_params)

    mz = np.zeros(len(roi))
    spint = mz.copy()

    # set subtract values to negative
    for k, r in enumerate(roi):
        sign = -np.ones(r.time.size)
        start_index, end_index = np.searchsorted(r.time, [params.start_time, params.end_time])
        sign[start_index:end_index] = 1
        mz[k] = np.nanmean(r.mz)
        spint[k] = np.nansum(r.spint * sign)

    # remove negative values
    pos_values = spint > 0
    mz = mz[pos_values]
    spint = spint[pos_values]

    # sort values
    sorted_index = np.argsort(mz)
    mz = mz[sorted_index]
    spint = spint[sorted_index]

    sp = MSSpectrum(mz=mz, int=spint, ms_level=params.ms_level)
    return sp


def _accumulate_spectra_profile(ms_data: MSData, params: AccumulateSpectraParameters) -> MSSpectrum:
    # The spectra are accumulated in two steps:
    #
    #  1.  iterate through scans to build a grid of m/z values for the
    #      accumulated spectra.
    #  2.  A second iteration is done to interpolate the intensity in each
    #      scan to the m/z grid and generate the accumulated spectrum.
    #
    #  This process is done in two steps to avoid storing the intensity
    #  values from each scan until the grid is built.

    # first iteration. Builds a grid of m/z values for the accumulated
    # spectrum. The grid is extended using new m/z values that appear
    # in each new scan
    accumulated_mz = _make_mz_filter(ms_data, MakeRoiParameters(**params.model_dump()))

    assert params.subtract_left_time is not None

    accumulated_sp = np.zeros_like(accumulated_mz)

    with ms_data.using_tmp_config(
        ms_level=params.ms_level, start_time=params.subtract_left_time, end_time=params.subtract_right_time
    ):
        for sp in ms_data:
            interpolator = interp1d(sp.mz, sp.int, fill_value=0.0)
            if (sp.time < params.start_time) or (sp.time > params.end_time):
                sign = -1
            else:
                sign = 1
            accumulated_sp += interpolator(accumulated_mz) * sign

    # set negative values that may result from subtraction to zero
    is_positive_sp = accumulated_sp > 0
    accumulated_mz = accumulated_mz[is_positive_sp]
    accumulated_sp = accumulated_sp[is_positive_sp]

    return MSSpectrum(mz=accumulated_mz, int=accumulated_sp, centroid=False)


class _RoiMaker:
    """Create and extends ROIs using spectrum data.

    Auxiliary class to make_roi

    Attributes
    ----------
    params : MakeRoiParameters
    mz_filter : sorted array
        m/z values used as a first filter for the values of spectra provided.
        m/z values in spectra with distances larger than `tolerance` are
        ignored.
    targeted : bool
        If ``True``, the mean of each ROI is updated after a new element is
        appended. Else, the mean when the ROI was created is used.

    """

    def __init__(
        self, sample: Sample, params: MakeRoiParameters, roi_seeds: np.ndarray | None, targeted: bool = False
    ):
        self.tmp_roi_list = _TempRoiList(update_mean=not targeted)
        self.valid_roi: deque["_TempRoi"] = deque()
        self.mz_filter = roi_seeds
        self.params = params
        self.sample = sample
        self.targeted = targeted

        if roi_seeds is not None:
            self.tmp_roi_list.initialize(roi_seeds)

    def feed_spectrum(self, sp: MSSpectrum):
        """Use a spectrum to extend and create ROIs."""
        if self.mz_filter is not None:
            sp = _filter_invalid_mz(self.mz_filter, sp, self.params.tolerance)

        if self.tmp_roi_list.roi:
            match = _match_mz(self.tmp_roi_list.mz_mean, sp, self.params.tolerance, self.params.multiple_match)
            self.tmp_roi_list.extend(match)

            if not self.targeted:
                self.tmp_roi_list.insert(match.no_match)
        else:
            if not self.targeted:
                self.tmp_roi_list.insert(sp)

    def clear_completed_roi(self):
        """Flag ROI as completed.

        Completed valid ROIs are stored  in the `roi` attribute. Invalid ROIs are cleared.

        """
        finished_mask = self.tmp_roi_list.missing_count > self.params.max_missing
        finished_roi_index = np.where(finished_mask)[0]
        valid_mask = finished_mask & (self.tmp_roi_list.max_int >= self.params.min_intensity)
        valid_mask &= self.tmp_roi_list.length >= self.params.min_length

        valid_index = np.where(valid_mask)[0]

        for i in valid_index:
            r = self.tmp_roi_list.get_roi_copy(i)
            self.valid_roi.append(r)

        self.tmp_roi_list.clear(finished_roi_index)

    def flag_as_completed(self):
        """Mark all ROis as completed."""
        self.tmp_roi_list.missing_count[:] = self.params.max_missing + 1

    def tmp_roi_to_roi(self, valid_scan: np.ndarray, time: np.ndarray, pad: int) -> list[MZTrace]:
        """Convert completed valid _TempRoi objects into LCTraces.

        :param valid_scan: scan values used to build the ROIs.
        :param time: acquisition time of each scan.


        """
        scan_to_index = {x.item(): k for k, x in enumerate(valid_scan)}
        valid_roi = list()
        while self.valid_roi:
            tmp = self.valid_roi.popleft()

            roi = tmp.to_mz_trace(time, valid_scan, self.params.pad, scan_to_index, self.sample)

            valid_roi.append(roi)
        return valid_roi


class _TempRoiList:
    """Container object of Temporary ROI.

    Auxiliary class used in make_roi.

    :param roi : list of ROI, sorted by mean m/z value.
    :param mz_mean: tracks the mean m/z value of each ROI.
    :param mz_sum: sum of m/z values in each ROI, used to update the mean.
    :param max_int: maximum intensity stored in each ROI.
    :param missing_count: number of times that a ROI has not been extended after calling `extend`.
        Each time that a ROI is extended the count is reset to 0.
    :param length: number of elements in each ROI.
    :param update_mean: If set to ``True``, the mean of each ROI is updated after a new element is
        appended. Else, the mean when the ROI was created is used.

    """

    def __init__(self, update_mean: bool = True):
        self.roi: list["_TempRoi"] = list()
        self.mz_mean = np.array([])
        self.mz_sum = np.array([])
        self.max_int = np.array([])
        self.missing_count = np.array([], dtype=int)
        self.length = np.array([], dtype=int)
        self.update_mean = update_mean

    def insert(self, sp: MSSpectrum):
        """Create new ROI and insert them in the list while keeping the order."""
        index = np.searchsorted(self.mz_mean, sp.mz)
        # update roi tracking values
        self.mz_mean = np.insert(self.mz_mean, index, sp.mz)
        self.mz_sum = np.insert(self.mz_sum, index, sp.mz)
        self.max_int = np.insert(self.max_int, index, sp.int)
        self.missing_count = np.insert(self.missing_count, index, np.zeros_like(index))
        self.length = np.insert(self.length, index, np.ones_like(index))

        # insert new roi
        new_roi = _create_roi_list(sp)
        offset = 0
        for i, roi in zip(index, new_roi):
            self.roi.insert(i + offset, roi)
            offset += 1

    def get_roi_copy(self, index: int) -> _TempRoi:
        """Create a deep copy of a temp ROI."""
        roi_copy = deepcopy(self.roi[index])
        roi_copy.mz_mean = self.mz_mean[index].item()
        return roi_copy

    def extend(self, match: SpectrumMatch):
        """Extend existing ROI with matching values."""
        index = match.index
        mz = match.match.mz
        sp = match.match.int
        for i, m, s in zip(index, mz, sp):
            self.roi[i].append(m, s, match.match.index)
        self.length[index] += 1
        if self.update_mean:
            self.mz_sum[index] += mz
            self.mz_mean[index] = self.mz_sum[index] / self.length[index]
        self.max_int[index] = np.maximum(self.max_int[index], sp)
        self.missing_count += 1
        self.missing_count[index] = 0

    def clear(self, index: np.ndarray):
        """Empty the m/z, intensity and scan values stored in each ROI.

        The mean value of each ROI is kept.

        :param index: indices of ROI to clear.

        """
        for i in index:
            self.roi[i].clear()

        self.mz_sum[index] = 0
        self.max_int[index] = 0
        self.missing_count[index] = 0
        self.length[index] = 0

    def initialize(self, mz: np.ndarray):
        init_sp = MSSpectrum(mz=mz, int=mz, index=0)
        self.insert(init_sp)
        self.clear(np.arange(mz.size))


class _TempRoi:
    """Stores data from a ROI.

    Auxiliary class used in make_roi.
    """

    def __init__(self):
        """Create a new empty Temporary ROI."""
        self.mz = deque()
        self.spint = deque()
        self.scan = deque()
        self.mz_mean = 0.0

    def append(self, mz: float, spint: float, scan: int):
        """Append new m/z, intensity and scan values."""
        self.mz.append(mz)
        self.spint.append(spint)
        self.scan.append(scan)

    def clear(self):
        """Empty the m/z, intensity and scan values stored."""
        self.mz = deque()
        self.spint = deque()
        self.scan = deque()

    def pad(self, n: int, valid_scan: np.ndarray):
        """Pad the ROI m/z and intensity with NaN.

        Values are padded only if the scan number allows it, that means if there
        are `n` points to the left or the right of the minimum and maximum scan
        number of the ROI in `valid_scans`.
        """
        first_scan = self.scan[0]
        last_scan = self.scan[-1]
        start, end = np.searchsorted(valid_scan, [first_scan, last_scan + 1])
        left_pad_index = max(0, start - n)
        n_left = start - left_pad_index
        right_pad_index = min(valid_scan.size, end + n)
        n_right = right_pad_index - end

        # left pad
        self.mz.extendleft([np.nan] * n_left)
        self.spint.extendleft([np.nan] * n_left)
        self.scan.extendleft(valid_scan[left_pad_index:start][::-1])

        # right pad
        self.mz.extend([np.nan] * n_right)
        self.spint.extend([np.nan] * n_right)
        self.scan.extend(valid_scan[end:right_pad_index])

    def to_mz_trace(
        self,
        rt: np.ndarray,
        scans: np.ndarray,
        pad: int,
        scan_to_index: dict[int, int],
        sample: Sample,
    ) -> MZTrace:
        """Convert to a m/z trace."""
        start_idx, end_idx = np.searchsorted(scans, [self.scan[0], self.scan[-1] + 1])
        start_pad_idx = max(0, start_idx - pad)
        end_pad_idx = min(scans.size, end_idx + pad)

        trace_scans = scans[start_pad_idx:end_pad_idx].copy()
        trace_rt = rt[start_pad_idx:end_pad_idx].copy()

        start_idx -= start_pad_idx
        end_idx -= start_pad_idx
        end_pad_idx -= start_pad_idx
        start_pad_idx = 0

        found_index = list()
        missing_index = list()
        found_scans = set(self.scan)
        for idx, scan in enumerate(trace_scans[start_idx:end_idx], start=start_idx):
            if scan in found_scans:
                found_index.append(idx)
            else:
                missing_index.append(idx)

        trace_mz = np.ones(shape=trace_scans.size, dtype=float) * self.mz_mean
        trace_mz[found_index] = self.mz

        trace_sp = np.zeros(shape=trace_scans.size, dtype=float)
        trace_sp[found_index] = self.spint
        if trace_scans.size > len(self.mz):
            trace_sp[missing_index] = np.interp(missing_index, found_index, self.spint)

        if start_pad_idx < start_idx:
            slope = (trace_sp[start_idx + 1] - trace_sp[start_idx]) / (trace_rt[start_idx + 1] - trace_rt[start_idx])
            intercept = trace_sp[start_idx]
            for idx in range(start_pad_idx, start_idx):
                trace_sp[idx] = max(0.0, (idx - start_idx) * slope + intercept)

        if end_idx < end_pad_idx:
            slope = (trace_sp[end_idx - 1] - trace_sp[end_idx - 2]) / (trace_rt[end_idx - 2] - trace_rt[end_idx - 1])
            intercept = trace_sp[end_idx - 2]
            for idx in range(end_idx, end_pad_idx):
                trace_sp[idx] = max(0.0, (idx - end_idx - 1) * slope + intercept)
        return MZTrace(sample=sample, time=trace_rt, mz=trace_mz, scan=trace_scans, spint=trace_sp)

    def convert_to_roi(self, rt: np.ndarray, scans: np.ndarray, sample: Sample) -> MZTrace:
        """Convert a TemporaryRoi into a ROI object.

        :param rt: acquisition times of each scan
        :param scans: sorted scan numbers used to build the ROIs.

        """
        # new arrays that include missing values. These new arrays may include
        # scans from other ms levels that must be removed
        first_scan = self.scan[0]
        last_scan = self.scan[-1]
        size = last_scan + 1 - first_scan
        mz_tmp = np.ones(size) * np.nan
        spint_tmp = mz_tmp.copy()

        # copy values from the ROI to the new arrays
        scan_index = np.array(self.scan) - self.scan[0]
        mz_tmp[scan_index] = self.mz
        spint_tmp[scan_index] = self.spint

        # remove scan numbers from other ms levels (i.e. scan numbers that are
        # not in the scans array)
        start_ind, end_ind = np.searchsorted(scans, [first_scan, last_scan + 1])
        scan_tmp = scans[start_ind:end_ind].copy()
        valid_index = scan_tmp - first_scan
        mz_tmp = mz_tmp[valid_index]
        spint_tmp = spint_tmp[valid_index]
        rt_tmp = rt[scan_tmp].copy()

        return MZTrace(time=rt_tmp, spint=spint_tmp, mz=mz_tmp, scan=scan_tmp, sample=sample)


def _make_mz_filter(data: MSData, params: MakeRoiParameters) -> FloatArray1D:
    """Create a list of m/z values to initialize ROI.

    Auxiliary function to make_roi.
    """
    with data.using_tmp_config(ms_level=params.ms_level, start_time=params.start_time, end_time=params.end_time):
        mz_seed = [sp.mz[sp.int > params.min_intensity] for sp in data]
    return np.unique(np.hstack(mz_seed))


def _filter_invalid_mz(valid_mz: np.ndarray, sp: MSSpectrum, tolerance: float) -> MSSpectrum:
    """Find values in the spectrum that are within tolerance with the m/z values in the seed.

    Auxiliary function to _RoiProcessor.extend

    :param valid_mz: sorted array of m/z values.
    :param mz: m/z values to filter.
    :param sp: intensity values associated to each m/z.
    :param tolerance: m/z tolerance

    """
    closest_index = find_closest(valid_mz, sp.mz)
    dmz = np.abs(valid_mz[closest_index] - sp.mz)
    match_mask = dmz <= tolerance  # type: np.ndarray
    return MSSpectrum(mz=sp.mz[match_mask], int=sp.int[match_mask], index=sp.index)


def _create_roi_list(sp: MSSpectrum) -> list[_TempRoi]:
    roi_list = list()
    for m, s in zip(sp.mz, sp.int):
        roi = _TempRoi()
        roi.append(m, s, sp.index)
        roi_list.append(roi)
    return roi_list


class SpectrumMatch(pydantic.BaseModel):
    """Return type for match_mz."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    index: IntArray1D
    match: MSSpectrum
    no_match: MSSpectrum


def _match_mz(mz1: FloatArray1D, sp2: MSSpectrum, tolerance: float, mode: str) -> SpectrumMatch:
    """Find values in a spectrum that fall within the specified tolerance in a list of m/z values.

    :param mz1: _RoiProcessor mz_mean
    :param sp2: intensity values associated to mz2
    :param tolerance: tolerance used to match values
    :param mode: {"closest", "reduce"}
        Behavior when more than one peak in mz2 matches with a given peak in
        mz1. If mode is `closest`, then the closest peak is assigned as a
        match and the others are assigned to no match. If mode is `merge`, then
        a unique mz and int value is generated using the average of the mz and
        the sum of the intensities.
    :return: a tuple consisting of: matching indices, matching mz values in the spectrum, sp values
        in the spectrum, non matching m/z values and non matching sp values.

    """
    closest_index = find_closest(mz1, sp2.mz)
    dmz = np.abs(mz1[closest_index] - sp2.mz)
    match_mask = dmz <= tolerance  # type: np.ndarray
    no_match_mask = ~match_mask
    match_index = closest_index[match_mask]

    # check multiple_matches
    match_unique, match_first, match_count = np.unique(match_index, return_counts=True, return_index=True)

    # set match values
    match_index = match_unique
    sp_match = sp2.int[match_mask]
    mz_match = sp2.mz[match_mask]

    # solve multiple matches
    multiple_match_mask = match_count > 1
    multiple_match_first = match_first[multiple_match_mask]

    if match_first.size > 0:
        multiple_match_count = match_count[multiple_match_mask]
        if mode == "reduce":
            for first, count in zip(multiple_match_first, multiple_match_count):
                # mz1 and mz2 are both sorted, the multiple matches are
                # consecutive
                mz_multiple_match = mz_match[first : (first + count)]
                sp_multiple_match = sp_match[first : (first + count)]
                mz_match[first] = np.mean(mz_multiple_match)
                sp_match[first] = np.sum(sp_multiple_match)
        elif mode == "closest":
            match_index_mz = np.where(match_mask)[0][match_first]
            multiple_match_index_mz = match_index_mz[multiple_match_mask]
            iterator = zip(multiple_match_index_mz, multiple_match_first, multiple_match_count)
            for mz2_index, first, count in iterator:
                closest = np.argmin(dmz[mz2_index : mz2_index + count])
                # flag all multiple matches as no match except the closest one
                no_match_mask[mz2_index : mz2_index + count] = True
                no_match_mask[mz2_index + closest] = False
                mz_match[first] = mz_match[first + closest]
                sp_match[first] = sp_match[first + closest]
        else:
            msg = "mode must be `closest` or `merge`"
            raise ValueError(msg)

    match_sp = MSSpectrum(mz=mz_match[match_first], int=sp_match[match_first], index=sp2.index)
    no_match_sp = MSSpectrum(mz=sp2.mz[no_match_mask], int=sp2.int[no_match_mask], index=sp2.index)

    return SpectrumMatch(index=match_index, match=match_sp, no_match=no_match_sp)


def _compute_roi_seeds(ms_data: MSData, params: MakeRoiParameters) -> FloatArray1D:
    """Create ROI seed values used for untargeted ROI extraction in make_roi.

    Aux function for make_roi.

    The algorithm is as follows:

    -   Create an array with all seed candidates by concatenating all m/z values
        with intensities greater than `min_intensity` into an array.
    -   Sort the candidates array.
    -   Combine close candidates into clusters. The standard deviation of each
        cluster is lower than the `tolerance` parameter.
    -   Compute seed values as the mean from each cluster.

    """
    seed_candidates = _create_seed_candidates(ms_data, params)
    seeds = _combine_seed_candidates(seed_candidates, params)
    return seeds


def _combine_seed_candidates(mz: np.ndarray, params: MakeRoiParameters) -> FloatArray1D:
    N = mz.size

    if N == 0:
        return np.array([], dtype=mz.dtype)

    mz_cumsum = np.cumsum(mz)
    mz2_cumsum = np.cumsum(mz**2)
    start, end = 0, 0
    seed_list = list()

    var_thresh = params.tolerance**2
    previous_mean = 0.0
    previous_var = 0.0
    while end <= N:
        slice_mean, slice_var = _compute_slice_stats(mz_cumsum, mz2_cumsum, start, end)

        if slice_var > var_thresh:
            # create a cluster for the data using the previous valid value
            # arrays are copied to avoid references to the concatenated mz, sp
            # and scans arrays also use the original data dtype
            seed_list.append(previous_mean)
            start = end - 1
        else:
            end += 1
        previous_mean, previous_var = slice_mean, slice_var

    if 0.0 <= previous_var < var_thresh:
        seed_list.append(previous_mean)

    return np.array(seed_list)


def _create_seed_candidates(ms_data: MSData, params: MakeRoiParameters) -> FloatArray1D:
    """Concatenate and sort spectra using m/z values."""
    mz_list = list()
    with ms_data.using_tmp_config(start_time=params.start_time, end_time=params.end_time):
        for sp in ms_data:
            mz_list.append(sp.mz[sp.int >= params.min_intensity])

    # Round m/z values to the 5th decimal place to reduce the number of unique values
    mz = np.round(np.hstack(mz_list), 5)
    return np.unique(mz)


def _compute_slice_stats(mz_cumsum: np.ndarray, mz2_cumsum: np.ndarray, start: int, end: int) -> tuple[float, float]:
    """Compute the population mean and variance of the slice mz[start:end]."""
    # This method uses the formula used for estimation of the variance in the
    # welford algorithm.
    # The variance of the slice x[start:end] can be computed as follows
    # S: std; xm: mean; i= start; j = end;
    # N = j - i
    # xm = (\sum_{i}^{j - 1} x_{i}) / N
    # S^{2} = \sum_{i}^{j} (x_{i} - xm)^{2} / N
    # S^{2} = (\sum_{i}^{j} x_{i}^{2} - N * xm^{2}) / N
    # S^{2} is computed using cumulative sums
    size = end - start
    # computes the sum in the slice mz[start:end]

    if size == 0:
        return 0.0, 0.0

    if start == 0:
        mz_start = 0.0
        mz2_start = 0.0
    else:
        mz_start = mz_cumsum[start - 1]
        mz2_start = mz2_cumsum[start - 1]

    mz_sum_slice = mz_cumsum[end - 1] - mz_start
    mz2_sum_slice = mz2_cumsum[end - 1] - mz2_start

    mz_mean = mz_sum_slice / size
    mz_var = (mz2_sum_slice - mz_mean**2 * size) / size

    return mz_mean, mz_var
